#==================================================================================
# The code is ported from Bill Schottstedaet's 'generators.scm' 
# file available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
import pysndlib.generators as gens
if cython.compiled:
    from cython.cimports.pysndlib import clm
    from cython.cimports.pysndlib import generators as gens

# from generators.scm 
# uses nrssb generator

# --------------- organish ---------------- #
@cython.ccall
def organish(beg, dur, freq, amp, fm_index, amp_env=None):
    start: cython.long = clm.seconds2samples(beg)
    i: cython.long = 0
    carriers = [None] * 3
    fmoscs = [None] * 3
    ampfs = [None] * 3
    pervib = clm.make_triangle_wave(5, clm.hz2radians(freq * .003))
    ranvib = clm.make_rand_interp(6, clm.hz2radians(freq * .002))
    resc = gens.make_nrssb(340, 1.0, 5, .5)
    resf = clm.make_env([0,0,.05,1,.1,0,dur,0], scaler=(amp * .05), duration=dur)
    stop: cython.long = start + clm.seconds2samples(dur)
    for i in range(3):
        frq = freq * 2**i
        index1 = clm.hz2radians((fm_index * frq * 5.0) / math.log(frq))
        index2 = clm.hz2radians((fm_index * frq * 3.0 * (8.5 - math.log(frq)) / (3.0 + (frq * 0.001))))
        index3 = clm.hz2radians((fm_index * frq * 4.0) / math.sqrt(frq))
        carriers[i] = clm.make_oscil(frq)

        fmoscs[i] = clm.make_polywave(frq, partials=[1, index1, 3, index2, 4, index3])
    if not amp_env:
        amp_env = [0,0,1,1,2,1,3,0]
    ampfs[0] = clm.make_env(amp_env, scaler=amp, duration=dur)
    ampfs[1] = clm.make_env([0,0,.04,1,.075, 0,dur,0], scaler=amp*.0125, duration=dur)
    ampfs[2] = clm.make_env([0,0,.02,1,.05,0,dur,0], scaler=amp*.025, duration=dur)
    
    for i in range(start, stop):
        vib = clm.triangle_wave(pervib) + clm.rand_interp(ranvib)
        res: cython.double = 0.0
        res += clm.env(resf) * gens.nrssb(resc, 0.0)
        res += clm.oscil(carriers[0], (vib+clm.polywave(fmoscs[0], vib)))
        res += clm.env(ampfs[0]) * clm.oscil(carriers[0], ((vib*2)+clm.polywave(fmoscs[0], (vib*2))))
        res += clm.env(ampfs[1]) * clm.oscil(carriers[1], ((vib*4)+clm.polywave(fmoscs[1], (vib*4))))
        clm.outa(i, res)


if __name__ == '__main__': 
    with clm.Sound(play=True,statistics=True):
        for i in range(10):
            organish(i*.3, .4, (100 + (50 * i)), .5, 1.0, [0,0,1,1,2,.5,3,.25,4,.125,10,0])



#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
if cython.compiled:
    from cython.cimports.pysndlib import clm


# --------------- wurley ---------------- #
#   ;; from Perry Cook's Rhodey.cpp
@cython.ccall
def wurley(beg, dur, freq, amp):
    osc0 = clm.make_oscil(freq)
    osc1 = clm.make_oscil(freq * 4.0)
    osc2 = clm.make_oscil(510.0)
    osc3 = clm.make_oscil(510.0)
    ampmod = clm.make_oscil(8.0)
    g0: cython.double = .5 * amp
    ampenv = clm.make_env([0,0,1,1,9,1,10,0], duration=dur)
    indenv = clm.make_env([0,0,.001,1,.15,0,max(dur, .16),0], duration=dur, scaler=.117)
    resenv = clm.make_env([0,0,.001,1,.25,0,max(dur, .26),0], duration=dur, scaler=(.5 * .307 * amp))
    st: cython.long = clm.seconds2samples(beg)
    nd: cython.long = clm.seconds2samples(beg + dur)
    
    i: cython.long = 0
    
    for i in range(st, nd):
        clm.outa(i, clm.env(ampenv) * 
            (1.0 + (.007 * clm.oscil(ampmod))) * 
            ((g0 * clm.oscil(osc0, (.307 * clm.oscil(osc1)))) + (clm.env(resenv) * clm.oscil(osc2, (clm.env(indenv) * clm.oscil(osc3))))))
            



#==================================================================================
# The code is an attempt at translation of Bill Schottstedaet's 'dsp.scm' 
# file available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

# TODO: things not translated yet - down_oct , Dolph-Chebyshev window, differentiator, IIR filters
# There are various fixes needed for items translated details in code

import math
from pathlib import Path
import cython
import pysndlib.clm as clm
# TODO: Figure out why I need to do this. 
from pysndlib.clm import ndarray2file

import numpy as np
from .env import scale_envelope
"hello there"

NEARLY_ZERO = 1.0e-10

# @cython.ccall
# def binomial(n, k):
#     if 0 <= k <= n:
#         a= 1
#         b=1
#         for t in range(1, min(k, n - k) + 1):
#             a *= n
#             b *= t
#             n -= 1
#         return a // b
#     else:
#         return 0

# --------------- src_duration ---------------- #


@cython.ccall
def src_duration(e) -> cython.double:
    """
      returns the new duration of a sound after using 'envelope' for time-varying sampling-rate conversion
    """
    length = len(e) - 2
    dur = 0.0
    all_x = e[length] - e[0]
    area = 0
    for i in range(0,length, 2):
        x0 = e[i]
        x1 = e[i+2]
        y0 = e[i+1]
        y1 = e[i+3]
        if abs(y0-y1) < .0001:
            area = (x1 - x0) / (y0 * all_x)
        else:
            if y0 == 0:
                raise ValueError("src can't use an evelope that has 0 as a y value")
            area = (math.log(y1 / y0) * (x1 - x0)) / ((y1-y0)*all_x)
        dur += abs(area)
    return dur
    
# --------------- src_fit_envelope ---------------- #

@cython.ccall
def src_fit_envelope(e, target_dur):
    return scale_envelope(e, src_duration(e) / target_dur)
    
    
    
# --------------- Dolph-Chebyshev window ---------------- #
# already in sndlib

    
    
# --------------- down_oct ---------------- #
# TODO: make this able to work with multi channel files
def down_oct(n, snd, chn, outname=None):
    if not clm.is_power_of_2(n):
        raise ValueError(f'n must be power of 2 not {n}')
    length = clm.get_length(snd)
    fftlen = int(math.floor(2**math.ceil(math.log(length, 2))))
    
    fftlen2 = fftlen // 2
    fft_1 = (n * fftlen) - 1
    fftscale = (1.0 / fftlen)
    rl1, sr = clm.file2ndarray(snd, chn, 0, fftlen )
    rl1 = rl1[0]
    im1 = np.zeros(fftlen)
    clm.mus_fft(rl1, im1, fftlen, 1)
    rl1 *= fftscale
    im1 *= fftscale
    rl2 = np.zeros(n * fftlen)
    im2 = np.zeros(n * fftlen)
    np.copyto(rl2[0:fftlen2], rl1[0:fftlen2])
    np.copyto(im2[0:fftlen2], im1[0:fftlen2])

    i: cython.long = 0
    j: cython.long = 0
    
    j = fft_1
    for k in range(fftlen-1, fftlen2):
        rl2[j] = rl1[k]
        im2[j] = im1[k]
        j -= 1
    clm.mus_fft(rl2, im2, (n * fftlen), -1)
    
    if outname is not None:
        ndarray2file(outname, rl2[:n*length], sr=sr)
        return outname
    else:
        ndarray2file(Path(snd).stem +'down_oct.wav', rl2[:n*length], sr=sr) 
        return Path(snd).stem +'_down_oct.wav'
    
    
# --------------- stretch-sound-via-dft ---------------- #
# Too slow

# def stretch_sound_via_dft(factor, snd, chn, outname=None):
#     n = clm.length(snd)
#     out_n = round(n*factor)
#     out_data = np.zeros(out_n)
#     fr = np.zeros(out_n)
#     freq = (np.pi*2) / n
#     in_data, sr = clm.file2ndarray(snd, chn, 0, n )
#     n2 = n // 2
#     in_data = in_data[0]
#     for i in range(n):
#         fr = i if i < n2 else ((out_n + i) - n - 1)
#         clm.edot_product(freq * (complex(0, -1) * i), in_data)
#     freq = ((np.pi*2) / out_n)
#     for i in range(out_n):
#         out_data[i] = clm.edot_product(freq * (complex(0, 1) * i), fr, n)   
#     if outname is not None:
#         ndarray2file(outname, out_data, sr=sr)
#         return outname
#     else:
#         ndarray2file(Path(snd).stem +'stretch_sound_via_dft.wav', out_data, sr=sr) 
#         return Path(snd).stem +'stretch_sound_via_dft.wav'


# --------------- freqdiv ---------------- #
@cython.ccall
def freqdiv(n, snd, chn, outname=None):
    data, sr = clm.file2ndarray(snd, chn, 0)
    size = clm.get_length(data)
    i: cython.long = 0
    k: cython.long = 0
    
    if n > 1:
        for i in range(0,size,n):
            val = data[0][i]
            stop = min(size, i+n)
            for k in range(i+1, stop):
                data[0][k] = val
        if outname is None:
            raise ValueError('missing outfile name')
    clm.ndarray2file(outname, data, size=size, sr=sr)


# -------- "adaptive saturation" -- an effect from sed_sed@my-dejanews.com

#-------- spike

#;;; -------- easily-fooled autocorrelation-based pitch tracker 

#;;; -------- chorus (doesn't always work and needs speedup)

# --------------- highpass ---------------- #

#-------- chordalize (comb filters to make a chord using chordalize-amount and chordalize-base)

# -------- zero-phase, rotate-phase

# -------- brighten-slightly

# ---FIR 

# -------- Hilbert transform


# freq in radians
@cython.ccall
def make_highpass(fc, length: cython.int =30):
    """makes an FIR highpass filter."""
    arrlen = 1 + (2 * length)
    arr = np.zeros(arrlen)
    for i in range(-length,length):
        k = i + length
        denom = math.pi * i
        num = -math.sin(fc * i)
        if i == 0:
            arr[k] = fc / math.pi
        else:
            arr[k] = (num / denom) * (.54 + (.46 * math.cos((i * math.pi) / length)))
    return clm.make_fir_filter(arrlen, arr)
    
    
@cython.ccall
def highpass(gen: clm.mus_any, insig: cython.double) -> cython.double:
    return clm.fir_filter(gen, insig)
# 
# --------------- lowpass ---------------- #
# freq in radians
@cython.ccall
def make_lowpass(fc, length: cython.int=30):
    """makes an FIR lowpass filter."""
    arrlen = 1 + (2 * length)
    arr = np.zeros(arrlen)
    for i in range(-length,length):
        k = i + length
        denom = math.pi * i
        num = math.sin(fc * i)
        if i == 0:
            arr[k] = fc / math.pi
        else:
            arr[k] = (num / denom) * (.54 + (.46 * math.cos((i * math.pi) / length)))
    return clm.make_fir_filter(arrlen, arr)

@cython.ccall
def lowpass(gen: clm.mus_any, insig: cython.double) -> cython.double:
    return clm.fir_filter(gen, insig)




# --------------- bandpass ---------------- #
# freq in radians
@cython.ccall
def make_bandpass(flo, fhi, length: cython.int =30):
    """makes an FIR bandpass filter."""
    arrlen = 1 + (2 * length)
    arr = np.zeros(arrlen)
    for i in range(-length,length):
        k = i + length
        denom = math.pi * i
        num = math.sin(fhi * i) - math.sin(flo * i)
        if i == 0:
            arr[k] = (fhi - flo) / math.pi
        else:
            arr[k] = (num / denom) * (.54 + (.46 * math.cos((i * math.pi) / length)))
    return clm.make_fir_filter(arrlen, arr)

@cython.ccall
def bandpass(gen: clm.mus_any, insig: cython.double) -> cython.double:
    return clm.fir_filter(gen, insig)

# --------------- bandstop ---------------- #
# freq in radians
@cython.ccall
def make_bandstop(flo, fhi, length: cython.int =30):
    """makes an FIR bandstop filter"""
    arrlen = 1 + (2 * length)
    arr = np.zeros(arrlen)
    for i in range(-length,length):
        k = i + length
        denom = math.pi * i
        num = math.sin(flo * i) - math.sin(fhi * i)
        if i == 0:
            arr[k] = (1.0 - (fhi - flo) / math.pi)
        else:
            arr[k] = (num / denom) * (.54 + (.46 * math.cos((i * math.pi) / length)))  
    return clm.make_fir_filter(arrlen, arr)
    
@cython.ccall
def bandstop(gen: clm.mus_any, insig: cython.double) -> cython.double:
    return clm.fir_filter(gen, insig)
# 
# 
# # --------------- TODO: differentiator ---------------- #
# 
# # --------------- IIR filters ---------------- #
# 
# --------------- Butterworth filters ---------------- #

@cython.ccall
def make_butter_high_pass(fq):
    """makes a Butterworth filter with high pass cutoff at 'freq'"""
    r = math.tan((math.pi * fq) / clm.get_srate())
    r2 = r * r
    c1 = 1.0 / (1.0 + (r * math.sqrt(2.0)) + r2)
    return clm.make_filter(3, [c1, (-2.0 * c1), c1], [0.0, 2.0*(r2-1.0)*c1, ((1.0 + r2 ) - (r * math.sqrt(2.0))) * c1])
#     
@cython.ccall
def make_butter_low_pass(fq):
    """makes a Butterworth filter with low pass cutoff at 'freq'"""
    r = 1.0 / math.tan((math.pi * fq) / clm.get_srate())
    r2 = r * r
    c1 = 1.0 / (1.0 + (r * math.sqrt(2.0)) + r2)
    return clm.make_filter(3, [c1, (2.0 * c1), c1], [0.0, 2.0*(1.0-r2)*c1, ((1.0 + r2 ) - (r * math.sqrt(2.0))) * c1])
    
@cython.ccall
def make_butter_band_pass(fq, bw):
    """makes a bandpass Butterworth filter with low edge at 'freq' and width 'band'"""
    c = 1.0 / math.tan((math.pi*bw) / clm.get_srate())
    d = 2.0 * math.cos((2.0 * math.pi * fq) / clm.get_srate())
    c1 = 1.0 / (1.0 + c)
    return clm.make_filter(3, [c1,0.0,-c1], [0., -c * d * c1, (c - 1.0) * c1])
 
@cython.ccall
def make_butter_band_reject(fq, bw):
    """makes a band-reject Butterworth filter with low edge at 'freq' and width 'band'"""
    c = 1.0 / math.tan((math.pi*bw) / clm.get_srate())
    c1 = 1.0 / (1.0 + c)
    c2 = c1 * -2.0 * math.cos((2.0*math.pi*fq) / clm.get_srate())
    return clm.make_filter(3, [c1,c2,c1], [0.0, c2, ((1.0 - c) * c1)])
    
@cython.ccall
def butter(gen: clm.mus_any, insig: cython.double) -> cython.double:
    return clm.filter(gen, insig) 
      

# --------------- Biquads ---------------- #

@cython.ccall
def make_biquad(a0,a1,a2,b1,b2):
    """returns a biquad filter (use with the CLM filter gen)"""
    return clm.make_filter(3, [a0,a1,a2], [0.0,b1,b2])

@cython.ccall
def make_local_biquad(a0, a1, a2, gamma, beta):
    return make_biquad(a0, a1, a2, (-2.0 * gamma), (2.0 * beta))

@cython.ccall
def make_iir_low_pass_2(fc, din):
    theta = (2 * math.pi * fc) / clm.get_srate()
    d = math.sin(theta) * ( (din or math.sqrt(2.0)) / 2.0)
    beta = .5 * ((1.0 - d) / (1.0 + d))
    gamma = (.5 + beta) * math.cos(theta)
    alpha = (.5 * ((.5 + beta) - gamma))
    return make_local_biquad(alpha, 2.0*alpha, alpha, gamma, beta)

@cython.ccall
def make_iir_band_pass_2(f1, f2):
    theta = (2 * math.pi * math.sqrt(f1*f2)) / clm.get_srate()
    t2 = math.tan(theta / (2 * (math.sqrt(f1*f2) / (f2-f1))))
    beta = .5 * ((1.0-t2)/(1.0+t2))
    gamma = (.5 * beta) * math.cos(theta)
    alpha = .5 - beta
    return make_local_biquad(alpha, 0.0, -alpha, gamma, beta)

@cython.ccall
def make_iir_band_stop_2(f1, f2):
    theta = (2 * math.pi * math.sqrt(f1*f2)) / clm.get_srate()
    t2 = math.tan(theta / (2 * (math.sqrt(f1*f2) / (f2-f1))))
    beta = .5 * ((1.0-t2)/(1.0+t2))
    gamma = (.5 * beta) * math.cos(theta)
    alpha = .5 + beta
    return make_local_biquad(alpha, -2.0*gamma, alpha, gamma, beta)

@cython.ccall
def biquad(gen: clm.mus_any, insig: cython.double) -> cython.double:
    clm.filter(gen, insig)




import math
import random
import cython
#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file  available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import numpy as np
import pysndlib.clm as clm
if cython.compiled:
    from cython.cimports.pysndlib import clm
# --------------- fm_insect ---------------- #
@cython.ccall
def fm_insect(startime, dur, frequency, 
                amplitude, 
                amp_env, 
                mod_freq, 
                mod_skew, 
                mod_freq_env, 
                mod_index, 
                mod_index_env, 
                fm_index, 
                fm_ratio, 
                degree=0.0, 
                distance=1.0,
                reverb_amount=.005):
                
    beg: cython.long = clm.seconds2samples(startime)
    end:cython.long  = clm.seconds2samples(startime + dur)
    loc = clm.make_locsig(degree, distance, reverb_amount)
    carrier = clm.make_oscil(frequency)
    fm1_osc = clm.make_oscil(mod_freq)
    fm2_osc = clm.make_oscil((fm_ratio * frequency))
    ampf = clm.make_env(amp_env, scaler=amplitude, duration=dur)
    indf = clm.make_env(mod_index_env, scaler=clm.hz2radians(mod_index), duration=dur)
    modfrqf = clm.make_env(mod_freq_env, scaler=clm.hz2radians(mod_skew), duration=dur)
    fm2_amp: cython.double = clm.hz2radians(fm_index * fm_ratio * frequency)
    
    i: cython.long = 0
    garble_in: cython.double = 0.
    
    for i in range(beg, end):
        garble_in = (clm.env(modfrqf) * clm.oscil(fm1_osc, clm.env(modfrqf)))
        clm.locsig(loc, i, (clm.env(ampf) * clm.oscil(carrier, (fm2_amp * clm.oscil(fm2_osc, garble_in)) + garble_in)))
        
        
if __name__ == '__main__': 
    LOCUST = [0,0,40,1,95,1,100,.5]
    BUG_HI = [0,1,25,.7,75,.78,100,1]
    AMP = [0,0,25,1,75,.7,100,0]

    with clm.Sound(play=True, statistics=True, channels=1):
        fm_insect(0,1.699,4142.627,.015,AMP,60,-16.707,LOCUST,500.866,BUG_HI,.346,.500)
        fm_insect(0.195,.233,4126.284,.030,AMP,60,-12.142,LOCUST,649.490,BUG_HI,.407,.500)
        fm_insect(0.217,2.057,3930.258,.045,AMP,60,-3.011,LOCUST,562.087,BUG_HI,.591,.500)
        fm_insect(2.100,1.500,900.627,.06,AMP,40,-16.707,LOCUST,300.866,BUG_HI,.346,.500)
        fm_insect(3.000,1.500,900.627,.06,AMP,40,-16.707,LOCUST,300.866,BUG_HI,.046,.500)
        fm_insect(3.450,1.500,900.627,.09,AMP,40,-16.707,LOCUST,300.866,BUG_HI,.006,.500)
        fm_insect(3.950,1.500,900.627,.12,AMP,40,-10.707,LOCUST,300.866,BUG_HI,.346,.500)
        fm_insect(4.300,1.500,900.627,.09,AMP,40,-20.707,LOCUST,300.866,BUG_HI,.246,.500)



#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file  available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================


import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
import pysndlib.env as env
if cython.compiled:
    from cython.cimports.pysndlib import clm

# --------------- exp_snd ---------------- #
# moved file arg from 1st arg
#   ;; granulate with envelopes on the expansion amount, segment envelope shape,
#   ;; segment length, hop length, and input file resampling rate

@cython.ccall
def is_pair(x): #seems only useful in this instrument
    return isinstance(x, list)

@cython.ccall
def exp_snd(beg, dur, file, amp, exp_amt=1.0, ramp=.4, seglen=.15, sr=1.0, hop=.05, ampenv=[0,0,.5,1,1,0]):

    max_seg_len = env.max_envelope(seglen) if is_pair(seglen) else seglen
    initial_seg_len = seglen[1] if is_pair(seglen) else seglen
    rampdata = ramp if is_pair(ramp) else [0, ramp, 1, ramp]
    max_out_hop = env.max_envelope(hop) if is_pair(hop) else hop
    initial_out_hop = hop[1] if is_pair(hop) else hop
    min_exp_amt = env.min_envelope(exp_amt) if is_pair(exp_amt) else exp_amt
    initial_exp_amt = exp_amt[1] if is_pair(exp_amt) else exp_amt

    if (env.min_envelope(rampdata) <= 0.0) or (env.max_envelope(rampdata) >= .5):
        raise RuntimeError(f'ramp argument to exp_snd must always be between 0.0 and .5: {ramp}.')

    st: cython.long = clm.seconds2samples(beg)
    nd: cython.long = clm.seconds2samples(beg + dur)
    f0 = clm.make_readin(file)

    expenv = clm.make_env( exp_amt if is_pair(exp_amt) else [0, exp_amt, 1, exp_amt], duration=dur)
    lenenv = clm.make_env( seglen if is_pair(seglen) else [0, seglen, 1, seglen], scaler=clm.default.srate, duration=dur)
    scaler_amp = ((.6 * .15) / max_seg_len) if max_seg_len > .15 else .6
    srenv = clm.make_env(sr if is_pair(sr) else [0, sr, 1, sr], duration=dur)
    rampenv = clm.make_env(rampdata, duration=dur)
    initial_ramp_time = ramp[1] if is_pair(ramp) else ramp
    max_in_hop = max_out_hop / min_exp_amt
    max_len = clm.seconds2samples(max(max_out_hop, max_in_hop) + max_seg_len)
    hopenv = clm.make_env( hop if is_pair(hop) else [0, hop, 1, hop], duration=dur)
    ampe = clm.make_env(ampenv, scaler= amp, duration=dur)
    exA = clm.make_granulate(f0, expansion=initial_exp_amt, max_size=max_len, ramp=initial_ramp_time, hop=initial_out_hop, length=initial_seg_len, scaler=scaler_amp)
    vol = clm.env(ampe)
    
    valA0: cython.double = vol * clm.granulate(exA)
    valA1: cython.double = vol * clm.granulate(exA)
    ex_samp: cython.double = 0.0
    next_samp: cython.double = 0.0
    
    i: cython.long = 0
    k: cython.long = 0
    
    for i in range(st, nd):
        sl: cython.double = clm.env(lenenv)
        vol: cython.double = clm.env(ampe)
        exA.mus_length = round(sl)
        exA.mus_ramp = math.floor(sl * clm.env(rampenv))
        exA.mus_frequency = clm.env(hopenv)
        exA.mus_increment = clm.env(expenv)
        next_samp += clm.env(srenv)
        
        if next_samp > (ex_samp + 1):
            samps = math.floor(next_samp - ex_samp)
            if samps > 2:
                for k in range(samps-2):
                    clm.granulate(exA)
            valA0 = (vol * clm.granulate(exA)) if samps >= 2 else valA1
            valA1 = vol * clm.granulate(exA)
            ex_samp += samps
            
            clm.outa(i, valA0 if next_samp == ex_samp else (valA0 + ((next_samp - ex_samp) * (valA1 - valA0)))) 

if __name__ == '__main__':       
    with clm.Sound(play=True,statistics=True):
        exp_snd(0, 3, '../examples/fyow.snd', 1, [0,1,1,3],   .4, .15, [0,2,1,.5], .05)




import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
import pysndlib.generators as gens
if cython.compiled:
    from cython.cimports.pysndlib import clm
    from cython.cimports.pysndlib import generators as gens

# uses nrcos generator


# --------------- lutish ---------------- #
@cython.ccall
def lutish(beg, dur, freq, amp):
    res1: cython.double = max(1, round(1000.0 / max(1.0, min(1000., freq))))
    maxind: cython.double = max(.01, min(.3, ((math.log(freq) - 3.5) / 8.0)))
    gen = gens.make_nrcos(frequency=freq*res1,n=max(1, (res1 - 2)))
    mod = clm.make_oscil(freq)
    start: cython.long = clm.seconds2samples(beg)
    stop: cython.long = clm.seconds2samples(beg + dur)
    indx = clm.make_env([0,maxind, 1, maxind*.25, max(dur, 2.0), 0.0], duration=dur)
    amplitude = clm.make_env([0,0,.01,1,.2,1,.5,.5,1,.25, max(dur, 2.0), 0.0], duration=dur, scaler =amp)
    i: cython.long = 0
    for i in range(start, stop):
        ind: cython.double = clm.env(indx)
        gen.mus_scaler = ind
        clm.outa(i, clm.env(amplitude) * gens.nrcos(gen, (ind * clm.oscil(mod))))



if __name__ == '__main__':   
    with clm.Sound(clipped=False, statistics=True, play=True):
        for i in range(10):
            lutish(i*.1, 2, (100 * (i+1)), .05)



#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file  available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
if cython.compiled:
    from cython.cimports.pysndlib import clm

# --------------- rhodey ---------------- #
@cython.ccall
#   ;; from Perry Cook's Rhodey.cpp
def rhodey(beg, dur, freq, amp, base=.5):
    osc0 = clm.make_oscil(freq)
    osc1 = clm.make_oscil(freq * .5)
    osc2 = clm.make_oscil(freq)
    osc3 = clm.make_oscil(freq * 15.0)
    ampenv1 = clm.make_env([0,0,.005, 1, dur, 0], base=base, duration=dur, scaler=(amp * .5))
    ampenv2 = clm.make_env([0,0,.001, 1, dur, 0], base=(1.5 * base), duration=dur, scaler=(amp * .5))
    ampenv3 = clm.make_env([0,0,.001, 1, .25, 0, max(dur, .26), 0], base=(4. * base), duration=dur, scaler= .109)
    st: cython.long = clm.seconds2samples(beg)
    nd: cython.long = clm.seconds2samples(beg + dur)
    
    i: cython.long = 0
        
    for i in range(st, nd):
        clm.outa(i, ( clm.env(ampenv1)*clm.oscil(osc0, (.535 * clm.oscil(osc1)))) + (clm.env (ampenv2) * clm.oscil(osc2, clm.env(ampenv3) * clm.oscil(osc3))))



#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file  available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
import pysndlib.env as env
if cython.compiled:
    from cython.cimports.pysndlib import clm

# --------------- reson ---------------- #
@cython.ccall
def reson(startime, dur, pitch, amp, numformants, indxfun, skewfun, pcskew, skewat, skewdc,
		      vibfreq, vibpc, ranvibfreq, ranvibpc, degree, distance, reverb_amount, data):
		      
    beg: cython.long = clm.seconds2samples(startime)
    end: cython.long = clm.seconds2samples(startime + dur)
    carriers = [None] * numformants
    modulator = clm.make_oscil(pitch)
    ampfs = [None] * numformants
    indfs = [None] * numformants
    c_rats = [None] * numformants
    totalamp = 0.0
    loc = clm.make_locsig(degree, distance, reverb_amount)
    pervib = clm.make_triangle_wave(vibfreq, clm.hz2radians(vibpc * pitch))
    ranvib = clm.make_rand_interp(ranvibfreq, clm.hz2radians(ranvibpc * pitch))
    frqe = clm.stretch_envelope(skewfun, 25, 100 * (skewat / dur), 75, 100 - (100 * (skewdc / dur)))
    frqf = clm.make_env(frqe, scaler=clm.hz2radians(pcskew * pitch), duration=dur)

    for k in range(0,numformants):
        totalamp += data[k][2]
        
    for k in range(0, numformants):
        frmdat = data[k]
        freq = frmdat[k]
        harm = round(freq / pitch)
        rfamp = frmdat[2]
        ampat = 100 * (frmdat[3] / dur)
        ampdc = 100 - (100 * (frmdat[4] / dur))
        dev0 = clm.hz2radians(frmdat[5] * freq)
        dev1 = clm.hz2radians(frmdat[6] * freq)
        indxat = 100 * (frmdat[7] / dur)
        indxdc = 100 - (100 * (frmdat[8] / dur))
        ampf = frmdat[0]
        rsamp = 1.0 - math.fabs(harm - (freq / pitch))
        cfq = pitch * harm
        if ampat == 0:
            ampat = 25
        if ampdc == 0:
            ampdc = 75
        if indxat == 0:
            indxat = 25
        if indxdc == 0:
            indxdc = 75
        indfs[k] = clm.make_env(env.stretch_envelope(indxfun, 25, indxat, 75, indxdc), duration=dur, scaler=(dev1 - dev0), offset=dev0)
        ampfs[k] = clm.make_env(env.stretch_envelope(ampf, 25, ampat, 75, ampdc), duration=dur, scaler=(rsamp * amp * rfamp) / totalamp)
        c_rats[k] = harm
        carriers[k] = clm.make_oscil(cfq)
        
    i: cython.long = 0
    j: cython.long = 0
    vib: cython.double = 0.0
    modsig: cython.double = 0.0
    outsum: cython.double
    
    if numformants == 2:
        e1: clm.mus_any = ampfs[0]
        e2: clm.mus_any = ampfs[1]
        c1: clm.mus_any = carriers[0]
        c2: clm.mus_any = carriers[1]
        i1: clm.mus_any = indfs[0]
        i2: clm.mus_any = indfs[1]
        r1: cython.double = c_rats[0]
        r2: cython.double = c_rats[1]
        
        for i in range(beg, end):
            vib = clm.env(frqf) + clm.triangle_wave(pervib) + clm.rand_interp(ranvib)
            modsig = clm.oscil(modulator, vib)
            clm.locsig(loc, i, (clm.env(e1) * clm.oscil(c1, (vib * r1) + (clm.env(i1) * modsig))) + (clm.env(e2) * clm.oscil(c2, (vib * r2) + (clm.env(i2) * modsig))))
        
    else:
        for i in range(beg, end):
            outsum = 0.0
            vib = clm.env(frqf) + clm.triangle_wave(pervib) + clm.rand_interp(ranvib)
            modsig = clm.oscil(modulator, vib)
            
            for j in range(0,numformants):
                outsum += clm.env(ampfs[j]) * clm.oscil(carriers[j], (vib * c_rats[j]) + (clm.env(indfs[j]) * modsig))
            
            clm.locsig(loc, i, outsum)       



#==================================================================================
# The code is ported from Bill Schottstedaet's 'generators.scm' 
# file  available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
import pysndlib.generators as gens
if cython.compiled:
    from cython.cimports.pysndlib import clm
    from cython.cimports.pysndlib import generators as gens

# from generators.scm 
# uses ercos generators

# --------------- ercoser ---------------- #
@cython.ccall
def ercoser(start, dur, freq, amp, r):
    beg: cython.long = clm.seconds2samples(start)
    end: cython.long = clm.seconds2samples(start + dur)
    gen = gens.make_ercos(freq, r)
    ampf = clm.make_env([0.0,0.0,.001, 1, .7, .5, 1.0, 0], duration=dur)
    t_env = clm.make_env([0,.1,1,2], duration=dur)
    
    for i in range(beg, end):
        r: cython.double  = clm.env(t_env)
        gen.cosh_t: cython.double = math.cosh(r)
        gen.osc.mus_data[0] = gen.cosh_t
        exp_t: cython.double = math.exp(-r)
        gen.offset = ((1.0 - exp_t) / (2.0 * exp_t))
        gen.scaler = (math.sinh(r) * gen.offset)
        clm.outa(i, amp*gens.ercos(gen)*clm.env(ampf))
#         
if __name__ == '__main__':          
    with clm.Sound(clipped=False, statistics=True, play=True):
        ercoser(0,1,100,.5,.1)  
        ercoser(.5,1,200,.5,.1)  
        ercoser(1,1,300,.5,.1)  
        ercoser(1.5,1,400,.5,.1)  



#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file  available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
if cython.compiled:
    from cython.cimports.pysndlib import clm

# --------------- nrev ---------------- #
# ;;; NREV (the most popular Samson box reverb)
# 
#   ;; reverb-factor controls the length of the decay -- it should not exceed (/ 1.0 .823)
#   ;; lp-coeff controls the strength of the low pass filter inserted in the feedback loop
#   ;; output-scale can be used to boost the reverb output
@cython.ccall
def is_even(n):
    return bool(n%2==0)

@cython.ccall
def next_prime(n):
    np=[]
    isprime=[]
    for i in range (n+1,n+200):
        np.append(i)
    for j in np:
        val_is_prime = True
        for x in range(2,j-1):
            if j % x == 0:
                val_is_prime = False
                break
        if val_is_prime:
            isprime.append(j)
    return min(isprime)

@clm.clm_reverb
def nrev(reverb_factor=1.09, lp_coeff=.7, volume=1.0, decay_time=1.):
    srscale = clm.get_srate() / 25641
    dly_len = [1433,1601,1867,2053,2251,2399,347,113,37,59,53,43,37,29,19] # TODO: Make this adjust to samplerate

    chans = clm.get_channels(clm.default.output)
    chan2 = chans > 1
    chan4 = chans == 4
    
    for i in range(0,15):
        val = math.floor(srscale * dly_len[i])
        if is_even(val):
            val += 1
        dly_len[i] = next_prime(val)
        
    length: cython.long = math.floor(clm.get_length(clm.default.reverb) + (clm.get_srate()*decay_time))
    comb1 = clm.make_comb(.822 * reverb_factor, dly_len[0])
    comb2 = clm.make_comb(.802 * reverb_factor, dly_len[1])
    comb3 = clm.make_comb(.733 * reverb_factor, dly_len[2])
    comb4 = clm.make_comb(.753 * reverb_factor, dly_len[3])
    comb5 = clm.make_comb(.753 * reverb_factor, dly_len[4])
    comb6 = clm.make_comb(.733 * reverb_factor, dly_len[5])
    low = clm.make_one_pole(lp_coeff, lp_coeff - 1.0)
    allpass1 = clm.make_all_pass(-.7000, .700, dly_len[6])
    allpass2 = clm.make_all_pass(-.7000, .700, dly_len[7])
    allpass3 = clm.make_all_pass(-.7000, .700, dly_len[8])
    allpass4 = clm.make_all_pass(-.7000, .700, dly_len[9])
    allpass5 = clm.make_all_pass(-.7000, .700, dly_len[11])
    allpass6 = clm.make_all_pass(-.7000, .700, dly_len[12]) if chan2 else None
    allpass7 = clm.make_all_pass(-.7000, .700, dly_len[13]) if chan4 else None
    allpass8 = clm.make_all_pass(-.7000, .700, dly_len[14]) if chan4 else None
    
    filts = []

    if not chan2:
        filts.append(allpass5)
    else:
        if not chan4:
            filts.extend([allpass5, allpass6])
        else:
            filts.extend([allpass5, allpass6, allpass7, allpass8])

    combs: clm.mus_any = clm.make_comb_bank([comb1, comb2, comb3, comb4, comb5, comb6])
    allpasses = clm.make_all_pass_bank([allpass1, allpass2, allpass3])
    
    i: cython.long = 0
    
    for i in range(length):
        clm.out_bank(filts, i, clm.all_pass(allpass4, clm.one_pole(low, clm.all_pass_bank(allpasses, clm.comb_bank(combs, volume * clm.ina(i, clm.default.reverb))))))
    
    



#==================================================================================
# The code is ported from Bill Schottstedaet's 'generators.scm' 
# file  available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
import pysndlib.generators as gens
if cython.compiled:
    from cython.cimports.pysndlib import clm
    from cython.cimports.pysndlib import generators as gens

# from generators.scm 
# uses r2k!cos generator

# --------------- pianoy1 ---------------- #

@cython.ccall
def pianoy1(start, dur, freq, amp, bounce_freq=5, bounce_amp=20):
    gen = gens.make_r2kfcos(freq, r=.5, k=3)
    gen1 = clm.make_oscil(bounce_freq)
    bouncef = clm.make_env([0,1,1,0], base=32, scaler=bounce_amp, duration=1.0)
    rf = clm.make_env([0,0,1,1,max(2.0,dur), 0], base=32, scaler=.1, offset=.25, duration=dur)
    ampf = clm.make_env([0,0,.01,1,.03,1,1,.15, max(2,dur), 0.0], base=32, scaler=amp, duration=dur)
    beg: cython.long = clm.seconds2samples(start)
    end: cython.long = clm.seconds2samples(start+dur)
    ln = clm.seconds2samples(dur)
    fv: np.ndarray = np.zeros(ln)
    i: cython.long = 0
    j: cython.long = 0
    for i in range(ln):
        fv[i] = clm.env(rf) + abs(clm.env(bouncef) * clm.oscil(gen1))
    
    for i in range(beg, end):
        gen.r = fv[j]
        clm.outa(i, clm.env(ampf) * gens.r2kfcos(gen))
        j += 1

if __name__ == '__main__': 
    with clm.Sound(statistics=True, play=True):
        pianoy1(0,4,200,.5, 1, .1)



#==================================================================================
# The code is ported from Bill Schottstedaet's 'generators.scm' 
# file available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
import pysndlib.generators as gens
if cython.compiled:
    from cython.cimports.pysndlib import clm
    from cython.cimports.pysndlib import generators as gens

# from generators.scm 
# uses make_rxykcos generator

# --------------- brassy ---------------- #
@cython.ccall
def brassy(start, dur, freq, amp, ampf, freqf, gliss):
    gen = clm.make_rxykcos(freq, 1, 0.0)
    beg = clm.seconds2samples(start)
    end = clm.seconds2samples(start+dur)
    amp_env = clm.make_env(ampf, duration=dur, scaler=amp)
    pitch_env = clm.make_env(freqf, scaler=gliss / freq, duration=dur)
    slant = clm.make_moving_average(clm.seconds2samples(.05))
    vib = clm.make_polywave(5, [1, clm.hz2radians(4.0)], clm.Polynomial.SECOND_KIND)
    harmfrq: cython.double = 0.0
    harmonic: cython.double = 0.0
    dist: cython.double = 0.0
    slant.mus_increment = clm.hz2radians(freq) * slant.mus_increment
    for i in range(beg, end):
        harmfrq = clm.env(pitch_env)
        harmonic = math.floor(harmfrq)
        dist = abs(harmfrq - harmonic)
        gen.mus_scaler = 20.0 * min(.1, dist, (1.0 - dist))
        clm.outa(i, clm.env(amp_env) * clm.rxykcos(gen, clm.moving_average(slant, harmonic) + clm.polywave(vib)))


if __name__ == '__main__': 

    with clm.Sound(clipped=False, statistics=True, play=True):
        brassy(0,4,50,.05,[0,0,1,1,10,1,11,0], [0,1,1,0], 1000)



#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
if cython.compiled:
    from cython.cimports.pysndlib import clm

# --------------- tubebell ---------------- #
# ;;; taken from Perry Cook's stkv1.tar.Z (Synthesis Toolkit), but I was
# ;;; in a bit of a hurry and may not have made slavishly accurate translations.
# ;;; Please let me know of any errors.

@cython.ccall
def tubebell(beg, dur, freq, amp, base=32.0):
    osc0 = clm.make_oscil(freq * .995)
    osc1 = clm.make_oscil(freq * 1.414 * 0.995)
    osc2 = clm.make_oscil(freq * 1.005)
    osc3 = clm.make_oscil(freq * 1.414)
    ampenv1 = clm.make_env([0,0,.005, 1, dur, 0], base=base, duration=dur, scaler=(amp * .5 * .707))
    ampenv2 = clm.make_env([0,0,.001, 1, dur, 0], base=(2 * base), duration=dur, scaler=(amp * .5))
    ampmod = clm.make_oscil(2.0)
    st: cython.long = clm.seconds2samples(beg)
    nd: cython.long = clm.seconds2samples(beg + dur)
    
    i: cython.long = 0
    
    for i in range(st, nd):
        clm.outa(i, ((.007 * clm.oscil(ampmod)) + .993) *
                ((clm.env(ampenv1) * clm.oscil(osc0, (.203 * clm.oscil(osc1)))) +
                 (clm.env(ampenv2) * clm.oscil(osc2, (.144 * clm.oscil(osc3))))))

if __name__ == '__main__':  
    import random

    with clm.Sound( play = True, statistics=True):
        freqs = [1174.659, 1318.510, 1396.913, 1567.982, 2093.005]
        for t in np.arange(0, 10, .4):
            tubebell(t, 1, random.choice(freqs), .5)    




from .attract import attract
from .bouncy import bouncy
from .brassy import brassy
from .cellon import cellon
from .clm_expsrc import clm_expsrc
from .ercoser import ercoser
from .exp_snd import exp_snd
from .fm_bell import fm_bell
from .fm_drum import fm_drum
from .fm_insect import fm_insect
from .fm_trumpet import fm_trumpet
from .fm_violin import fm_violin
from .fm_voice import fm_voice
from .fofins import fofins
from .glassy import glassy
from .gong import gong
from .gran_synth import gran_synth
from .hammondoid import hammondoid
from .jc_reverb import jc_reverb
from .jl_reverb import jl_reverb
from .lbj_piano import lbj_piano
from .lutish import lutish
from .machine1 import machine1
from .metal import metal
from .nkssber import nkssber
from .nrev import nrev
from .oboish import oboish
from .organish import organish
from .pianoy import pianoy
from .pianoy1 import pianoy1
from .pianoy2 import pianoy2
from .pluck import pluck
from .pqw import pqw
from .resflt import resflt
from .reson import reson
from .rhodey import rhodey
from .scratch import scratch
from .spectra import spectra
from .stereo_flute import stereo_flute
from .stringy import stringy
from .touch_tone import touch_tone
from .two_tab import two_tab
from .tubebell import tubebell
from .wurley import wurley
from .za import za
from .zc import zc
from .zn import zn





#! /Users/toddingalls/Developer/Python/venvs/pysndlib-venv/bin/python
# Examples from https://ccrma.stanford.edu/software/snd/snd/sndclm.html#oscildoc
# showing basic examples of built-in generators
import math
import os

from random import uniform
import numpy as np
from pysndlib.sndlib import *
from pysndlib.clm import *
from pysndlib.jcrev import jc_reverb
from pysndlib.musx import *
import functools




# with Sound(play=True, statistics=True):
#     gen = make_oscil(440.0)
#     for i in range(44100):
#         outa(i,  oscil(gen))
# 
#
# with Sound(play=True, statistics=True):
#     gen = make_oscil(440.0)
#     ampf = make_env([0., 0., 0.01, 1.0, 0.25, 0.1, 1, 0], scaler=.5, length=44100)
#     for i in range(44100):
#         outa(i,  env(ampf)*oscil(gen))
#     
#      
# with Sound(play=True, statistics=True):
#     gen = make_table_lookup(440.0, wave=partials2wave([1., .5, 2, .5]))
#     for i in range(44100):
#         outa(i, .5 * table_lookup(gen))
# 
# with Sound(play=True, statistics=True):
#     gen = make_polywave(440, partials=[1., .5, 2, .5])
#     for i in range(44100):
#         outa(i, .5 * polywave(gen))
# 
# 
# with Sound(play=True, statistics=True):
#     gen = make_triangle_wave(440.0)
#     for i in range(44100):
#         outa(i, .5 * triangle_wave(gen))
# 
# with Sound(play=True, statistics=True):
#     gen = make_ncos(440.0, 10);
#     for i in range(44100):
#         outa(i, .5 * ncos(gen))
# 
# 
# with Sound(play=True, statistics=True):
#     gen = make_nrxycos(440.0,n=10)
#     for i in range(44100):
#         outa(i, .5 * nrxycos(gen))
# 
# 
# with Sound(play=True, statistics=True):
#     v = np.zeros(64, dtype=np.double)
#     g = make_ncos(400,10)
#     g.mus_phase = -.5 * math.pi
#     for i in range(64):
#         v[i] = ncos(g)
#     gen = make_wave_train(440, wave=v)    
#     for i in range(44100):
#         outa(i, .5 * wave_train(gen))
#         
#         
# 
# with Sound(channels=2, play=True, statistics=True):
#     ran1 = make_rand(5.0, hz2radians(220.0))
#     ran2 = make_rand_interp(5.0, hz2radians(220.0))
#     osc1 = make_oscil(440.0)
#     osc2 = make_oscil(1320.0)
#     for i in range(88200):
#         outa(i, 0.5 * oscil(osc1, rand(ran1)))
#         outb(i, 0.5 * oscil(osc2, rand_interp(ran2)))
#         
#!         
# with Sound(play=True, statistics=True):
#     flt = make_two_pole(1000.0, 0.999)
#     ran1 = make_rand(10000.0, 0.002)
#     for i in range(44100):
#         outa(i, 0.5 * two_pole(flt, rand(ran1)))
#         
#         
# with Sound(play=True, statistics=True):
#     flt = make_two_pole(1000.0, .999)
#     ran1 = make_rand(10000.0, .002)
#     for i in range(44100):
#         outa(i, .5 * two_pole(flt, rand(ran1)))
# 
# with Sound(play=True, statistics=True):
#     flt = make_firmant(1000.0, 0.999)
#     ran1 = make_rand(10000.0, 5.0)
#     for i in range(44100):
#         outa(i, 0.5 * firmant(flt, rand(ran1)))
# 
#     
# with Sound(play=True, statistics=True):
#     flt = make_iir_filter(3, [0.0, -1.978, 0.998])
#     ran1 = make_rand(10000.0, 0.002)
#     for i in range(44100):
#         outa(i, 0.5 * iir_filter(flt, rand(ran1)))
# 
# with Sound(play=True, statistics=True):
#     dly = make_delay(seconds2samples(0.5))
#     osc1 = make_oscil(440.0)
#     osc2 = make_oscil(660.0)
#     for i in range(44100):
#         outa(i, 0.5 * (oscil(osc1) + delay(dly, oscil(osc2))))        
# 
# with Sound(play=True, statistics=True):
#     cmb = make_comb(0.4, seconds2samples(0.4))
#     osc = make_oscil(440.0)
#     ampf = make_env([0.0, 0.0, 1.0, 1.0, 2.0, 1.0, 3.0, 0.0], length=4410)
#     
#     for i in range(88200):
#         outa(i, 0.5 * (comb(cmb, env(ampf) * oscil(osc))))
# 
# with Sound(play=True, statistics=True):
#     alp = make_all_pass(-0.4, 0.4, seconds2samples(0.4))
#     osc = make_oscil(440.0)
#     ampf = make_env([0.0, 0.0, 1.0, 1.0, 2.0, 1.0, 3.0, 0.0], length=4410)
#     for i in range(88200):
#         outa(i, 0.5 * (all_pass(alp, env(ampf) * oscil(osc))))
#         
# 
# with Sound(play=True, statistics=True):
#     avg = make_moving_average(4410)
#     osc = make_oscil(440.0)
#     stop = 44100 - 4410
#     #avg.print_cache()
#     for i in range(stop):
#         val = oscil(osc)
#         outa(i, val * moving_average(avg, abs(val))) 
#     for i in range(stop, 44100):
#         outa(i, oscil(osc) * moving_average(avg, 0.0))
#         
#     
# with Sound(play=True, srate=22050, statistics=True):    
#     rd = make_readin("oboe.snd");
#     length = 2 * mus_sound_framples("oboe.snd")
#     sr = make_src(rd, .5);
#     
#     for i in range(length):
#         outa(i, src(sr))
#     
# # slightly different as don't have samples  function. 
# with Sound(play=True, statistics=True):    
#     flt, _ = file2ndarray('oboe.snd', channel=0, beg=0, dur=clm_length('pistol.snd'))
#     cnv = make_convolve(make_readin('pistol.snd'), flt[0])
#     for i in range(88200):
#         outa(i, .25 * convolve(cnv))
#     
# with Sound(play=True, statistics=True):    
#     grn = make_granulate(make_readin('oboe.snd'), 2.)
#     for i in range(44100):
#         outa(i, granulate(grn))
#         
#         

# with Sound(play=True, statistics=True):    
#     sweep = make_env([0,0,1,1], scaler=hz2radians(440.0), length=44100)
#     osc = make_oscil(440.0)
#     @inputcallback
#     def cb(g, d):
#         return .2 * oscil(osc, env(sweep))
# 
#     grn = make_granulate(cb, length=.5)
#     for i in range(88200):
#         outa(i, granulate(grn))
#     
#         
# with Sound(play=True, statistics=True):    
#     pv = make_phase_vocoder(make_readin("oboe.snd"), pitch=1.4)
#     for i in range(44100):
#         outa(i, phase_vocoder(pv))    
# 
# with Sound(play=True, statistics=True):    
#     fm = make_asymmetric_fm(440.0, 0.0, 0.9, 0.5)
#     for i in range(44100):
#         outa(i, 0.5 * asymmetric_fm(fm, 1.0))
#         
# with Sound(play=True, statistics=True):    
#     reader = make_readin("oboe.snd")
#     for i in range(44100):
#         outa(i, readin(reader))
# 
#        
# with Sound(play=True, statistics=True):    
#     infile = make_file2sample('oboe.snd')
#     for i in range(44100):
#         out_any(i, in_any(i, 0, infile), 0)
# 
# 
# with Sound(channels=2, play=True, statistics=True):    
#     loc = make_locsig(60.0)
#     osc = make_oscil(440.0)
#     for i in range(44100):
#         locsig(loc, i, .5 * oscil(osc))
#         
# 
# # # -----------------------------
# # # some other examples from sndclm
# # 
# def simp(beg, dur, freq, amp, envelope):
#     os = make_oscil(freq)
#     amp_env = make_env(envelope, duration=dur, scaler=amp)
#     start = seconds2samples(beg)
#     end = seconds2samples(beg + dur)
#     for i in range(start, end):
#         outa(i, env(amp_env) * oscil(os))
#         
# with Sound(play=True, statistics=True):
#     simp(0,2,440,.1,[0,0,.1,1.0,1.0,0.0])
#     
#     
# def simp(start, end, freq, amp, frq_env):
#     os = make_oscil(freq)
#     frqe = make_env(frq_env, length= (end+1) - start, scaler=hz2radians(freq))
#     for i in range(start, end):
#         outa(i, amp * oscil(os, env(frqe)))
#          
# with Sound(play=True, statistics=True):
#     simp(0,10000,440,.1,[0, 0, 1, 1])
#     
#     
# def simple_fm(beg, dur, freq, amp, mc_ratio, idx, amp_env=None, index_env=None):
#     start = seconds2samples(beg)
#     end = start + seconds2samples(dur)
#     cr = make_oscil(freq)
#     md = make_oscil(freq * mc_ratio)
#     fm_index = hz2radians(idx * mc_ratio * freq)
#     ampf = make_env(amp_env or [0,0,.5,1,1,0],scaler=amp, duration=dur)
#     indf = make_env(index_env or [0,0,.5,1,1,0], scaler=fm_index, duration=dur)
#     for i in range(start, end):
#         outa(i, env(ampf) * oscil(cr, env(indf) * oscil(md)))  
#     
# with Sound(play=True, statistics=True):
#     simple_fm(0, 1, 440, .1, 2, 1.0)
#     
#     
#     
# with Sound(play=True, statistics=True):
#     peaks = [23,0.0051914,32,0.0090310,63,0.0623477,123,0.1210755,185,0.1971876,
#             209,0.0033631,247,0.5797809,309,1.0000000,370,0.1713255,432,0.9351965,
#             481,0.0369873,495,0.1335089,518,0.0148626,558,0.1178001,617,0.6353443,
#             629,0.1462804,661,0.0208941,680,0.1739281,701,0.0260423,742,0.1203807,
#             760,0.0070301,803,0.0272111,865,0.0418878,926,0.0090197,992,0.0098687,
#             1174,0.00444,1298,0.0039722,2223,0.0033486,2409,0.0083675,2472,0.0100995,
#             2508,0.004262,2533,0.0216248,2580,0.0047732,2596,0.0088663,2612,0.0040592,
#             2657,0.005971,2679,0.0032541,2712,0.0048836,2761,0.0050938,2780,0.0098877,
#             2824,0.003421,2842,0.0134356,2857,0.0050194,2904,0.0147466,2966,0.0338878,
#             3015,0.004832,3027,0.0095497,3040,0.0041434,3092,0.0044802,3151,0.0038269,
#             3460,0.003633,3585,0.0050849,4880,0.0042301,5121,0.0037906,5136,0.0048349,
#             5158,0.004336,5192,0.0037841,5200,0.0038025,5229,0.0035555,5356,0.0045781,
#             5430,0.003687,5450,0.0055170,5462,0.0057821,5660,0.0041789,5673,0.0044932,
#             5695,0.007370,5748,0.0031716,5776,0.0037921,5800,0.0062308,5838,0.0034629,
#             5865,0.005942,5917,0.0032254,6237,0.0046164,6360,0.0034708,6420,0.0044593,
#             6552,0.005939,6569,0.0034665,6752,0.0041965,7211,0.0039695,7446,0.0031611,
#             7468,0.003330,7482,0.0046322,8013,0.0034398,8102,0.0031590,8121,0.0031972,
#             8169,0.003345,8186,0.0037020,8476,0.0035857,8796,0.0036703,8927,0.0042374,
#             9388,0.003173,9443,0.0035844,9469,0.0053484,9527,0.0049137,9739,0.0032365,
#             9853,0.004297,10481,0.0036424,10490,0.0033786,10606,0.0031366]
#     
#     length = len(peaks) // 2
#     dur = 10
#     oscs = [None] * length
#     amps = [None] * length
#     ramps = [None] * length
#     freqs = [None] * length
#     vib = make_rand_interp(50, hz2radians(.01))
#     ampf = make_env([0,0,1,1,10,1,11,0], duration=dur, scaler=.1)
#     samps = seconds2samples(dur)
#     for i in range(length):
#         freqs[i] = peaks[i*2]
#         oscs[i] = make_oscil(freqs[i], uniform(0, math.pi))
#         amps[i] = peaks[1+(2*i)]
#         ramps[i] = make_rand_interp(1.0 + (i * (20.0 / length)), 
#                                     (.1 + (i * (3.0 / length)))*amps[i])
#     
#     for i in range(samps):
#         sm = 0.0
#         fm = rand_interp(vib)
#         for k in range(length):
#             sm += (amps[k] + rand_interp(ramps[k])) * oscil(oscs[k], freqs[k]*fm)
#         
#         outa(i, env(ampf) * sm)
#     
#     
# 
# with Sound(channels=2, play=True, statistics=True):
#     dur = 2.0
#     samps = seconds2samples(dur)
#     pitch = 1000
#     modpitch = 100
#     pm_index = 4.0
#     fm_index = hz2radians(4.0 * modpitch)
#     car1 = make_oscil(pitch)
#     mod1 = make_oscil(modpitch)
#     car2 = make_oscil(pitch)
#     mod2 = make_oscil(modpitch)
#     frqf = make_env([0,0,1,1], duration=dur)
#     ampf = make_env([0,0,1,1,20,1,21,0], duration=dur, scaler=.5)
#     for i in range(samps):
#         frq = env(frqf)
#         rfrq = hz2radians(frq)
#         amp = env(ampf)
#         outa(i, amp * (oscil(car1, (rfrq*pitch) + (fm_index * (frq + 1) * oscil(mod1, (rfrq * modpitch))))))
#         outb(i, amp * (oscil(car2, (rfrq*pitch) , pm_index * oscil(mod2, rfrq * modpitch))))
#         
#         
#     car1 = make_oscil(pitch)
#     mod1 = make_oscil(modpitch)
#     car2 = make_oscil(pitch)
#     mod2 = make_oscil(modpitch)
#     frqf = make_env([0,0,1,1], duration=dur)
#     ampf = make_env([0,0,1,1,20,1,21,0], duration=dur, scaler=.5)
#     for i in range(samps):
#         frq = env(frqf)
#         rfrq = hz2radians(frq)
#         amp = env(ampf)
#         outa(i, amp * (oscil(car1, (rfrq*pitch) + (fm_index * oscil(mod1, (rfrq * modpitch))))))
#         outb(i, amp * (oscil(car2, (rfrq*pitch) , (pm_index / (frq + 1)) * oscil(mod2, rfrq * modpitch))))
# 
#       
#     
#     
# with Sound(play=True, statistics=False):
#     phases98 = [0.000000, -0.183194, 0.674802, 1.163820, -0.147489, 1.666302, 0.367236, 0.494059, 0.191339,
#                 0.714980, 1.719816, 0.382307, 1.017937, 0.548019, 0.342322, 1.541035, 0.966484, 0.936993,
#                 -0.115147, 1.638513, 1.644277, 0.036575, 1.852586, 1.211701, 1.300475, 1.231282, 0.026079,
#                 0.393108, 1.208123, 1.645585, -0.152499, 0.274978, 1.281084, 1.674451, 1.147440,0.906901,
#                 1.137155, 1.467770, 0.851985, 0.437992, 0.762219, -0.417594, 1.884062, 1.725160, -0.230688,
# 		        0.764342, 0.565472, 0.612443, 0.222826, -0.016453, 1.527577, -0.045196, 0.585089, 0.031829,
# 		        0.486579, 0.557276, -0.040985, 1.257633, 1.345950, 0.061737, 0.281650, -0.231535, 0.620583,
# 		        0.504202, 0.817304, -0.010580, 0.584809, 1.234045, 0.840674, 1.222939, 0.685333, 1.651765,
# 		        0.299738, 1.890117, 0.740013, 0.044764, 1.547307, 0.169892, 1.452239, 0.352220, 0.122254,
# 		        1.524772, 1.183705, 0.507801, 1.419950, 0.851259, 0.008092, 1.483245, 0.608598, 0.212267,
# 		        0.545906, 0.255277, 1.784889, 0.270552, 1.164997, -0.083981, 0.200818, 1.204088]
# 		        
#     freq = 10.0
#     dur = 5.0
#     n = 98
# 	
#     samps = math.floor(dur * 44100)
#     onedivn = 1.0 / n
#     freqs = np.zeros(n)
#     phases = np.zeros(n)
#     phases.fill(math.pi*.5)
#     for i in range(n):
#         off = (math.pi * (.5 - phases98[i])) / dur / 44100
#         h = hz2radians(freq*(i+1))
#         freqs[i] = h + off
#     
#     ob = make_oscil_bank(freqs, phases)
#     for i in range(1000): #get rid of the distracting initial click
#         nada = oscil_bank(ob) # added in this assignment because otherwise was printing output
#     for k in range(samps):
#         outa(k, onedivn * oscil_bank(ob))
# 
# 
#
# def mapenv(beg, dur, frq, amp, en):
#     start = seconds2samples(beg)
#     end = start + seconds2samples(dur)
#     osc = make_oscil(frq)
#     zv = make_env(en, 1.0, dur)
#     for i in range(start, end):
#         zval = env(zv)
#         outa(i, amp * math.sin(.5*math.pi*zval*zval*zval)*oscil(osc))
#     
# with Sound(play=True):
#     mapenv(0, 1, 440, .5, [0,0,50,1,75,0,86,.5,100,0])   
#     
#     
# 
# with Sound(play=True, statistics=True):
#     e = make_env([0,0,1,1,2,.25,3,1,4,0], duration=.5)
#     for i in range(44100):
#           outa(i, env_any(e, lambda y : y * y))
#     
# 
#     
# def sine_env(e):
#     return env_any(e, lambda y : .5 * (1.0 + math.sin((-.5*math.pi) + (math.pi*y))))
#     
# with Sound(play=True, statistics=True):
#     e = make_env([0,0,1,1,2,.25,3,1,4,0], duration=.5)
#     for i in range(44100):
#         outa(i, sine_env(e))
# 
# 
# with Sound(play=True, statistics=True):
#     e = make_pulsed_env([0,0,1,1,2,0], .01, 1)
#     frq = make_env([0,0,1,1], duration=1.0, scaler=hz2radians(50))
#     for i in range(44100):
#         outa(i, .5 * pulsed_env(e, env(frq)))
#     
# 
# with Sound(play=True):
#     wav = make_polyshape(frequency=500, partials=[1, .5, 2, .3, 3, .2])
#     for i in range(40000):
#         outa(i, 1. * polyshape(wav))
#     
#     
# 
# with Sound(play=True):
#     harms = np.zeros(10)
#     k = 1
#     for i in range(0,10,2):
#         harms[i] = k
#         harms[i+1] = (1.0 / math.sqrt(k))
#         k+= 3 
#     gen = make_polywave(200, harms)
#     ampf = make_env([0,0,1,1,10,1,11,0], duration=1.0, scaler=.5)
#     for i in range(44100):
#         outa(i, env(ampf) * polywave(gen) * .5)
#     
#     
# def pqw(start, dur, spacing, carrier, partials):
#     spacing_cos = make_oscil(spacing, math.pi / 2.0)
#     spacing_sin = make_oscil(spacing)
#     carrier_cos = make_oscil(carrier, math.pi / 2.0)
#     carrier_sin = make_oscil(carrier)
#     sin_coeffs = partials2polynomial(partials, Polynomial.SECOND_KIND)
#     cos_coeffs = partials2polynomial(partials, Polynomial.FIRST_KIND)
#     beg = seconds2samples(start)
#     end = beg + seconds2samples(dur)
#     for i in range(beg, end):
#         ax = oscil(spacing_cos)
#         outa(i, (oscil(carrier_sin) * oscil(spacing_sin) * polynomial(sin_coeffs, ax)) -
#                 (oscil(carrier_cos) * polynomial(cos_coeffs, ax)))
#     
#     
# with Sound(play=True, statistics=True):
#     pqw(0,1,200.0,1000.0, [2,.2,3,.3,6,.5])
#
#     
# with Sound(play=True, statistics=True):
#     modulator = make_polyshape(100, partials=[0,.4,1,.4,2,.1,3,.05,4,.05])    
#     carrier = make_oscil(1000.)
#     for i in range(20000*2):
#         outa(i, .5 * oscil(carrier) * polyshape(modulator))
#     
# 
# with Sound(play=True, statistics=True):
#     dur = 7
#     samps = seconds2samples(dur)
#     coeffs = [0.0, 0.5, .25, .125, .125]
#     x = 0.0
#     incr = hz2radians(100)
#     ampf = make_env([0,0,1,1,10,1,11,0], duration=dur, scaler=.5)
#     harmf = make_env([0, .125, 1, .25], duration=dur)
#     for i in range(samps):
#         harm = env(harmf)
#         coeffs[3] = harm
#         coeffs[4] = .25 - harm
#         outa(i, env(ampf) * chebyshev_t_sum( x, coeffs))
#         x += incr
#     
#     
# with Sound(play=True, statistics=True):
#     gen = make_polyshape(100, partials=[11,1,20,1])
#     ampf = make_env([0,0,1,1,20,1,21,0], scaler=.4, length=88200)
#     indf = make_env([0, 0, 1, 1, 1.1, 1], length=88200)
#     for i in range(88200):
#         outa(i, env(ampf)*polyshape(gen, env(indf)))

# 
# with Sound(play=True, statistics=True):
#     gen = make_polyshape(1000.0, partials=[1, .25, 2, .25, 3, .125, 4, .125, 5, .25])
#     indf = make_env([0,0,1,1,2,0], duration = 2.0)
#     ampf = make_env([0,0,1,1,2,1,3,0], duration=2.0)
#     mx = make_moving_max(256)
#     samps = seconds2samples(2.0)
#     for i in range(samps):
#         val = polyshape(gen, env(indf))
#         outa(i, (env(ampf)*val) / max(.001, moving_max(mx, val)))
#     
# 
# with Sound(play=True, statistics=True):
#     pcoeffs = partials2polynomial([5,1])
#     gen1 = make_oscil(100.)
#     gen2 = make_oscil(2000.0)
#     for i in range(44100):
#         outa(i, polynomial(pcoeffs, .5 * (oscil(gen1) + oscil(gen2))))
#     
# with Sound(play=True, channels=2, statistics=True):
#     dur = 2.0
#     samps = seconds2samples(dur)
#     p1 = make_polywave(800, [1,.1,2,.3,3,.4,5,.2])
#     p2 = make_polywave(400, [1,.1,2,.3,3,.4,5,.2])
#     interpf = make_env([0,0,1,1], duration=dur)
#     p3 = partials2polynomial([1,.1,2,.3,3,.4,5,.2])
#     g1 = make_oscil(800)
#     g2 = make_oscil(400)
#     ampf = make_env([0,0,1,1,10,1,11,0], duration=dur)
#     
#     for i in range(samps):
#         interp = env(interpf)
#         amp = env(ampf)
#         #chan A: interpolate from one spectrum to the next directly
#         outa(i, amp * ((interp * polywave(p1)) + ((1.0 - interp) * polywave(p2))))
#         #chan B: interpolate inside the sum of Tns!
#         outb(i, amp*(polynomial(p3, (interp * oscil(g1)) + ((1.0-interp) * oscil(g2)))))
# 
#
# 
# with Sound(play=True, statistics=True):
#     duty_factor = .25
#     p_on = make_pulse_train(100, .5)
#     p_off = make_pulse_train(100, -.5, (2 * math.pi * (1 - duty_factor)))
#     sm = 0.0
#     for i in range(44100*4):
#         sm += pulse_train(p_on) + pulse_train(p_off)
#         outa(i, sm)
# 
# 
# def simple_soc(beg, dur, freq, amp):
#     os = make_ncos(freq, 10)
#     start = seconds2samples(beg)
#     end = start + seconds2samples(dur)
#     for i in range(start, end):
#         outa(i, amp*ncos(os))
# 
# 
# with Sound(play=True):
#     simple_soc(0,1,100,.9)
# 
# 
# 
#   
# with Sound(play=True):
#     gen1 = make_nrxycos(400, 1, 15, .95)
#     indr = make_env([0,-1,1,1], length=80000, scaler=.9999)
#     for i in range(80000):
#         gen1.mus_scaler = env(indr)
#         outa(i, .5 * nrxycos(gen1, 0.0))

# 
# 
# def shift_pitch(beg, dur, file, freq, order=40):
#     st = seconds2samples(beg)
#     nd = st + seconds2samples(dur)
#     gen = make_ssb_am(freq, order)
#     rd = make_readin(file)
#     for i in range(st, nd):
#         outa(i, ssb_am(gen, readin(rd)))
# 
# with Sound(play=True):
#     shift_pitch(0, 3, 'oboe.snd', 1108.0)
#     
# 
# with Sound(play=True,  statistics=True):
#     gen1 = make_ssb_am(750, 40)
#     gen2 = make_ssb_am(700, 40)
#     rd = make_readin('oboe.snd')
#     for i in range(44100):
#         outa(i, .5*(ssb_am(gen1, readin(rd)) + ssb_am(gen2, readin(rd))))
    
    
    
    
# # 
# # 
# # 
# # 
# def fofins(beg, dur, frq, amp, vib, f0, a0, f1, a1, f2, a2, ve=[0,1,100,1], ae=[0,0,25,1,75,1,100,0]):
#     start = seconds2samples(beg)
#     end = start + seconds2samples(dur)
#     ampf = make_env(ae, scaler=amp, duration=dur)
#     frq0 = hz2radians(f0)
#     frq1 = hz2radians(f1)
#     frq2 = hz2radians(f2)
#     foflen = 100 if default.srate == 22050 else 200
#     vibr = make_oscil(6)
#     vibenv = make_env(ve , scaler=vib, duration=dur)
#     win_freq = (2 * math.pi) / foflen
#     foftab = np.zeros(foflen)
# 
#     wt0 = make_wave_train(frq, foftab)
# 
#     for i in range(foflen):
#         #this is not the pulse shape used by B&R
#         foftab[i] = ((a0 * math.sin(i*frq0)) + (a1 * math.sin(i*frq1)) +
#                     (a2 * math.sin(i*frq2))) * .5 * (1.0 - math.cos(i * win_freq))
#     
#     for i in range(start, end):
#         outa(i, env(ampf) * wave_train(wt0, env(vibenv) * oscil(vibr)))  
# 
# with Sound(play=True, statistics=True):
#     fofins(0, 1, 270, .2, .001, 730, .6, 1090, .3, 2440, .1) # "Ahh"

# 
# with Sound( play = True , statistics=True): #one of JC's favorite demos
#     fofins(0,4,270,.2,0.005,730,.6,1090,.3,2440,.1,[0,0,40,0,75,.2,100,1],
#             [0,0,.5,1,3,.5,10,.2,20,.1,50,.1,60,.2,85,1,100,0])
#     fofins(0,4,(6/5 * 540),.2,0.005,730,.6,1090,.3,2440,.1,[0,0,40,0,75,.2,100,1],
#             [0,0,.5,.5,3,.25,6,.1,10,.1,50,.1,60,.2,85,1,100,0])
#     fofins(0,4,135,.2,0.005,730,.6,1090,.3,2440,.1,[0,0,40,0,75,.2,100,1],
#             [0,0,1,3,3,1,6,.2,10,.1,50,.1,60,.2,85,1,100,0])
#test.wav: : maxamp: [0.7691921] [1740] compute time: 0.16470929 seconds. 
#         #    
#             
#             
#      
# TODO: THIS NEEDS to be fixed
# def when(start, duration, start_freq, end_freq, grain_file):
#     beg = seconds2samples(start)
#     length = seconds2samples(duration)
#     end = beg + length
#     grain_dur = clm_length(grain_file)
#     print(grain_dur)
#     frqf = make_env([0,0,1,1], scaler = hz2radians(end_freq - start_freq), duration=duration)
#     click_track = make_pulse_train(start_freq)
#     grain_size =  grain_dur
#     print(grain_size)
#     grain_arr, sr = file2ndarray(grain_file, 0, 0, grain_size) #different than scheme version
#     grains = make_wave_train(start_freq, grain_arr[0]) # TODO: make_wave_train just take size arg
#     original_grain = np.copy(grain_arr[0])
#     ampf = make_env([0,1,1,0], scaler=.7, offset=.3, duration=duration, base=3.0)
#      #grain = grains.mus_data this needs work TODO:
#     grain = grains.mus_data
# #     
#     for i in range(beg, end):
#         gliss = env(frqf)
#         outa(i, env(ampf) * wave_train(grains, gliss))
#         click = pulse_train(click_track, gliss)
#         if click > 0.0:
#             scaler = max(.1, (1.0 * ((i - beg) / length)))
#             comb_len = 32
#             c1 = make_comb(scaler, comb_len)
#             c2 = make_comb(scaler, math.floor(comb_len * .75))
#             c3 = make_comb(scaler, math.floor(comb_len * 1.25))
#             for k in range(0, grain_size):
#                 x = original_grain[k]
#                 #print(x)
#                 grain[k] = .25 * (comb(c1, x) + comb(c2, x) + comb(c3, x))
# 
# 
# with Sound(play=True):
#     when(0, 4, 2.0, 8.0, 'flute_trill_1.wav')
# # 
# # 
# # 
# def fractal(start, duration, m, x, amp):
#     # use formula of M J Feigenbaum
#     beg = seconds2samples(start)
#     end = beg + seconds2samples(duration)
#     for i in range(beg, end):
#         outa(i, amp * x)
#         x = (1.0 - (m * x * x))
# 
# with Sound(play=True, statistics=True):
#     fractal(0, 1, .5, 0, .5)
#     
# with Sound():
#     fractal(0, 1, 1.5, .20, .2)
# # 
# # 
# def attract(beg, dur, amp, c):
#     # by James McCartney, from CMJ vol 21 no 3 p 6
#     st = seconds2samples(beg)
#     nd = st + seconds2samples(dur)
#     a = .2
#     b = .2
#     dt = .04
#     scale = (.5 * amp) / c
#     x1 = 0.0
#     x = -1.
#     y = 0.0
#     z = 0.0
#     for i in range(st,nd):
#         x1 = x - (dt * (y+z))
#         y += dt * (x + (a * y))
#         z += dt * ((b+(x*z)) - (c * z))
#         x = x1
#         outa(i, (scale*x))
#         
#             
# with Sound(play=True, statistics=True):
#     attract(0, 2, .5, 7.4)
#     attract(2, 2, .5, 4.19)
#     attract(4, 2, .5, 6.37)
# #     
# #         
# #         
# def test_filter(flt):
#     osc = make_oscil()
#     samps = seconds2samples(.5)
#     ramp = make_env([0,0,1,1], scaler=hz2radians(samps), length=samps)
#     with Sound():
#         for i in range(0, samps):
#             outa(i, flt(oscil(osc, env(ramp))))    
#             
# test_filter(make_one_zero(.5,.5))  
# test_filter(make_one_pole(.1,-.9))      
# test_filter(make_two_pole(.1,.1,.9))  
# test_filter(make_two_zero(.5,.2,.3))  
# 
# 
# 
# def echo(beg, dur, scaler, secs, file):
#     dly = make_delay(seconds2samples(secs))
#     rd = make_readin(file)
#     for i in range(beg,dur):
#         inval = rd()
#         outa(i, inval + delay(dly, scaler * (tap(dly) + inval)))
#         
# with Sound(play=True, statistics=True):
#     echo(0, 60000, .5, 1.0, 'pistol.snd')
# #     
# # 
# # 
# # 
# # # another version in clm_ins.py
# def zc(time, dur, freq, amp, length1, length2, feedback):
#     beg = seconds2samples(time)
#     end = seconds2samples(time + dur)
#     s = make_pulse_train(freq)
#     d0 = make_comb(feedback, size=length1, max_size=(max(length1, length2)))
#     aenv = make_env([0,0,.1,1,.9,1,1,0], scaler=amp, duration=dur)
#     zenv = make_env([0,0,1,1], scaler=(length2-length1), base=12, duration=dur)
#     for i in range(beg, end):
#         outa(i, env(aenv) * comb(d0, pulse_train(s), env(zenv)))
#         
#         
# with Sound(play=True, statistics=True):
#     zc(0,3,100,.1,20,100,.5)
#     zc(3.5,3,100,.1,90,100,.95)

# 
# # !!
# TODO: fix mus_data
# def fir_plus_comb(beg, dur, freq, amp, size):
#     start = seconds2samples(beg)
#     end = start + seconds2samples(dur)
#     dly = make_comb(feedback=.9, size=size)
#     flt = make_fir_filter(order=size, xcoeffs=dly.mus_data)
#     r = make_rand(freq)
#     for i in range(start, end):
#         outa(i, amp * fir_filter(flt, comb(dly, rand(r))))
#         
# 
# with Sound(play=True):
#     fir_plus_comb(0,2,10000, .001, 200)
#     fir_plus_comb(2,2,1000, .0005, 400)
#     fir_plus_comb(4,2,3000, .001, 300)
#     fir_plus_comb(6,2,3000, .0005, 1000)  #!
# 
# 
# #callable for fucntion callback tests?
# 
# def flux(start, file, frequency, combs0, combs1, scaler=.99, comb_len=32):
#     beg = seconds2samples(start)
#     end = beg + clm_length(file)
#     num_combs0 = len(combs0)
#     num_combs1 = len(combs1)
#     cmbs0 = [None] * num_combs0
#     cmbs1 = [None] * num_combs1
#     cmbs0_scl = 1 / num_combs0
#     cmbs1_scl = 1 / num_combs1
#     osc = make_oscil(frequency)
#     rd = make_readin(file)
#     for k in range(num_combs0):
#         cmbs0[k] = make_comb(scaler, math.floor(comb_len * combs0[k]))
#     for k in range(num_combs1):
#         cmbs1[k] = make_comb(scaler, math.floor(comb_len * combs1[k]))
# 
#     nc0 = make_comb_bank(cmbs0)
#     nc1 = make_comb_bank(cmbs1)
#     
#     for i in range(beg, end):
#         interp = oscil(osc)
#         x = readin(rd)
#         outa(i, (interp * comb_bank(nc0,x)) + ((1.0 - interp) * comb_bank(nc1, x)))
#         
# with Sound(play=True, scaled_to=.5, clipped=False, statistics=True):
#     flux(0, 'oboe.snd', 10.0, [1.0, 1.25, 1.5], [1.0, 1.333, 1.6])
#     flux(2, 'flute_trill_1.wav', 4.0, [1.0, 1.25, 1.5], [1.0, 1.333, 1.6, 2.0, 3.0])
#     flux(4, 'flute_trill_1.wav', 1.0, [1.0, 1.25, 1.5], [1.0, 1.333, 1.6, 2.0, 3.0], .995, 20)
#     flux(6, 'flute_trill_1.wav', 10.0, [1.0, 1.25, 1.5], [1.0, 1.333, 1.6, 2.0, 3.0], .99, 10)
#     flux(8, 'flute_trill_1.wav', 10.0, [2.0], [1.0, 1.333, 1.6, 2.0, 3.0], .99, 120)
#     flux(10, 'fyow.snd', .5, [1.0, 2.0, 1.5], [1.0, 1.333, 1.6, 2.0, 3.0], .99, 120)
# 
# 
# 
# def simple_src(start, duration, amp, srt, srt_env, filename):
#     senv = make_env(srt_env, duration=duration)
#     beg = seconds2samples(start)
#     end = beg + seconds2samples(duration)
#     src_gen = make_src(make_readin(filename), srate=srt)
#     for i in range(beg, end):
#         outa(i, amp*src(src_gen, env(senv)))
#         
# with Sound(play=True):
#     simple_src(0, 4, 1.0, .5, [0,1,1,2], 'oboe.snd')
# 
# 
# def srcer(start, duration, amp, srt, fmamp, fmfreq, filename):
#     os = make_oscil(fmfreq)
#     beg = seconds2samples(start)
#     end = beg + seconds2samples(duration)
#     src_gen = make_src(make_readin(filename), srate=srt)
#     for i in range(beg, end):
#         outa(i, amp*src(src_gen, fmamp*oscil(os)))
#         
#         
# with Sound(play=True): srcer(0, 2, 1.0, 1, .3, 20, 'fyow.snd')
# with Sound(play=True): srcer(0, 25, 10.0, .01, 1, 10, 'fyow.snd')
# with Sound(play=True): srcer(0, 2, 1.0, .9, .05, 60, 'oboe.snd')
# with Sound(play=True): srcer(0, 2, 1.0, 1.0, .5, 124, 'oboe.snd')
# with Sound(play=True): srcer(0, 2, 10.0, .01, .2, 8, 'oboe.snd')
# with Sound(play=True): srcer(0, 2, 1.0, 1, 3, 20, 'oboe.snd')


# 
# def src_change(filename, start, duration, file_start, srcenv):
#     beg = seconds2samples(start)
#     end = beg + seconds2samples(duration)
#     loc = seconds2samples(file_start)
#     e = make_env(srcenv, duration=duration)
#     inp = make_file2sample(filename)
#     #wish python lambdas were better
#     def reader(g, d): 
#         nonlocal loc
#         loc += d
#         val =  ina(loc, inp)
#         return val
#     
#     src_gen = make_src(reader, srate=0.0) #Note function define in make_src not put in call to src
#     
#     for i in range(beg, end):
#         outa(i, src(src_gen, env(e)))
#     
# with Sound(play=True):
#     src_change('pistol.snd', 0, 2, 0, [0,.5, 1, -1.5])
#    
# def convins(beg, dur, filt, file, size=128):
#     start = seconds2samples(beg)
#     end = start + seconds2samples(dur)
#     ff = make_convolve(make_readin(file), filt, fft_size=size)
#     for i in range(start, end):
#         outa(i, convolve(ff))
#         
# with Sound(play=True):
#     convins(0, 2, [1.0, .5, .25, .125], 'oboe.snd')
#     
# with Sound(play=True):
#     flt = np.zeros(100000, dtype=np.double)
#     flt[0] = 1.0
#     flt[40000] = .3
#     convins(0, 2, flt, 'pistol.snd')
# 
# 
# # make an ir 
# with Sound('noiseverb.aif', play=True):
#     dur = 44100*3
#     e = make_env([0,1,1,0], length=dur, base=32)
#     r = make_rand(default.srate)
#     for i in range(0,dur):
#         outa(i, env(e)*rand(r))
# 
# # made convolve_files take some additional args 
# # so something like convolve-files->aifc is not needed
# # def convolve_files(file1, file2, maxamp: Optional[float]=1., outputfile='test.aif', sample_type=CLM.sample_type, header_type=CLM.header_type):
# # sndlay just convenvience 
# # plays sound with whatevern CLM.player is 
# import subprocess
# def sndplay(file):
#     subprocess.run([default.player,file])
# convolve_files('pistol.snd', 'noiseverb.aif', outputfile='pistol_conv.aif')
# sndplay('pistol_conv.aif')
# convolve_files('flute_trill_1.wav', 'noiseverb.aif', outputfile='flute_trill_1_conv.aif')
# 
# 
# def granulate_sound(file, beg, dur=None, orig_beg=0.0, exp_amt=1.0):
#     f_srate = clm_srate(file)
#     f = make_readin(file, start=round(f_srate*orig_beg))
#     st = seconds2samples(beg)
#     new_dur = dur or (mus_sound_duration(file) - orig_beg)
#     exA = make_granulate(f, expansion=exp_amt)
#     nd = st + seconds2samples(new_dur)
#     for i in range(st, nd):
#         outa(i, granulate(exA))
#         
# with Sound(play=True, statistics=True):
#     granulate_sound('flute_trill_1.wav', 0, 3.0, 0., 2.0)
#     
#     
# import cython
# def grev(beg, dur, exp_amt, file, file_beg):
#     fil = make_file2sample(file)
#     ctr = file_beg
#     # in example reader function is called when granulate
#     # function run. this is less efficient with how i have implemented callbacks
#     # than defining in make_granulate. something to look at in future
#     #TODO Add ability to change callback functionin granulate function
#     def reader(gen, d):
#         nonlocal fil
#         nonlocal ctr
#         inval = file2sample(fil, ctr, 0)
#         if ctr > 0:
#             ctr -= 1
#         return inval
#     exA = make_granulate(reader, exp_amt)
#     for i in range(beg, beg+dur):
#         outa(i, granulate(exA))
#     
# with Sound(play=True, statistics=True):
#     grev(0, 100000, 2.0, 'pistol.snd', 40000)
#     
# #another way of doing this in python would be using a generator 
# def read_backwards(fil, ctr):
#     loc = ctr
#     while True:
#         inval = file2sample(fil, loc)
#         if loc > 0:
#             loc -= 1
#         yield inval
#         
#         
#         
# def grev(beg, dur, exp_amt, file, file_beg):
#     fil = make_file2sample(file)
#     ctr = file_beg
#     reader = read_backwards(fil, ctr)
#     exA = make_granulate(lambda d : next(reader, 0.0) , exp_amt) # in this example direction is ignored
#     
#     for i in range(beg, beg+dur):
#         outa(i, granulate(exA))
# #             
# with Sound(play=True):
#     grev(0, 100000, 2.0, 'flute_trill_1.wav', 40000)    
#     
# # 
# 
# 
# 
# with Sound(play=True, statistics= True):
#     forward = True
#     rd = make_readin('flute_trill_1.wav')
#     #         
#     def reader(d): 
#         return readin(rd)
#     def grain_edit(g):
#         forward = True
#         grain = g.mus_data
#         length = g.mus_length
#         if forward:
#             forward = False
#             tmp = np.flip(grain)
#             g.mus_data = tmp# note the use of copy to here
#         return length
#     
#     grn = make_granulate(reader, edit=grain_edit, hop=.1, length=.3)  
#     for i in range(0,rd.mus_length):
#         outa(i, granulate(grn))

# 
# 
# ## This is different than the sndclm examples
# ## I am having difficulty with the resynthesize callback
# ## but this example uses the edit callback to filter 
# ## out amps below .1 
# # 
# import cython
# def pvedit(gen):
#     length = gen.mus_length
#     pvamps = phase_vocoder_amps(gen)
#     tmp = pvamps
#     np.copyto(pvamps, tmp)
#     return length
# 
# with Sound(statistics=True, play=True):
#     pv = make_phase_vocoder(make_readin('flute_trill_1.wav'), 512, 4, 128, 1.0, edit=pvedit)
#     for i in range(0, 44100):
#         outa(i, phase_vocoder(pv))

# # 
# 
# 

# def pvedit(gen):
#     length = gen.mus_length
#     print(length)
#     amps = phase_vocoder_amp_increments(gen)
#     tmp = np.where(amps < 0.25, 0, amps)
#     np.copyto(amps, tmp)
#     return length
# 
# 
# with Sound(statistics=True, play=True):
#     pv = make_phase_vocoder(make_readin("goodness.aiff"), 512, 4, 128, 1.0, edit=pvedit)
#     for i in range(0, get_length(("goodness.aiff"))):
#         outa(i, phase_vocoder(pv))
#         
#         
#       











import math
from pysndlib import *
import numpy as np

#==================================================================================
# The code is an attempt at translation of Bill Schottstedaet's 'analog-filter.scm' 
# file available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================


# def analog2digital(n, num, den, fz):
#     g = 1.0
#     Q = 1.0
#     wc = math.tan(math.pi * fz)
#     c = np.zeros(2 * n, dtype=np.double)
    
    






#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
if cython.compiled:
    from cython.cimports.pysndlib import clm

# --------------- stereo_flute ---------------- #
# TODO: Am getting strange artifact at beginning 
@cython.ccall
def stereo_flute(start,dur,freq,flow,
                flow_envelope = [0,1,100,1], 
                decay=.01, 
                noise=0.0356, 
                embouchure_size=0.5,
                fbk_scl1=.5, 
                fbk_scl2=.55, 
                offset_pos=0.764264,
                out_scl=1.0, 
                a0=.7, 
                b1=-.3, 
                vib_rate=5, 
                vib_amount=.03, 
                ran_rate=5, 
                ran_amount=0.03):
                    
    period_samples: cython.double = math.floor(clm.default.srate / freq)
    embouchure_samples: cython.double = math.floor(embouchure_size * period_samples)
    current_excitation: cython.double = 0.0
    current_difference: cython.double = 0.0
    current_flow: cython.double = 0.0
    out_sig: cython.double = 0.0
    tap_sig: cython.double = 0.0
    previous_out_sig: cython.double = 0.0
    previous_tap_sig: cython.double = 0.0
    dc_blocked_a: cython.double = 0.0
    dc_blocked_b: cython.double = 0.0
    previous_dc_blocked_a: cython.double = 0.0
    previous_dc_blocked_b: cython.double = 0.0
    delay_sig: cython.double = 0.0
    emb_sig: cython.double = 0.0
    beg: cython.long = clm.seconds2samples(start)
    end: cython.long = clm.seconds2samples(start + dur)
    flowf = clm.make_env(flow_envelope, scaler=flow, duration=dur - decay)
    periodic_vibrato = clm.make_oscil(vib_rate)
    random_vibrato: clm.mus_any  = clm.make_rand_interp(ran_rate, ran_amount)
    breath  = clm.make_rand(clm.default.srate/2, amplitude=noise)
    embouchure  = clm.make_delay(embouchure_samples, initial_element=0.0)
    bore = clm.make_delay(period_samples)
    offset = math.floor(period_samples * offset_pos)
    reflection_lowpass_filter = clm.make_one_pole(a0, b1)

    i: cython.long = 0

    for i in range(beg, end):
        
        delay_sig = clm.delay(bore, out_sig)
        emb_sig = clm.delay(embouchure, current_difference)
        current_flow = (vib_amount * clm.oscil(periodic_vibrato)) + clm.rand_interp(random_vibrato) + clm.env(flowf)
        current_difference = current_flow + (current_flow * clm.rand(breath) + (fbk_scl1 * delay_sig))
        current_excitation = emb_sig - (emb_sig*emb_sig*emb_sig)
        out_sig = clm.one_pole(reflection_lowpass_filter, (current_excitation + (fbk_scl2 * delay_sig)))
        tap_sig = clm.tap(bore, offset)
        dc_blocked_a = (out_sig + (.995 * previous_dc_blocked_a)) - previous_out_sig
        dc_blocked_b = (tap_sig + (.995 * previous_dc_blocked_b)) - previous_tap_sig
        clm.outa(i, (out_scl * dc_blocked_a))          
        clm.outb(i, (out_scl * dc_blocked_b)) 
        previous_out_sig = out_sig
        previous_dc_blocked_a = dc_blocked_a
        previous_tap_sig = tap_sig
        previous_dc_blocked_b = dc_blocked_b       
                    
                    
if __name__ == '__main__':  
    with clm.Sound( play = True, statistics=True, channels=2 ):
        stereo_flute(0,3,220,.55, flow_envelope=[0,0,25,1,75,1,100,0], ran_amount=.03)
        stereo_flute(3.5,3,220,.55, flow_envelope=[0,0,25,1,75,1,100,0], offset_pos=.5)     



#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
if cython.compiled:
    from cython.cimports.pysndlib import clm

# --------------- za ---------------- #
@cython.ccall
def za(time, dur, freq, amp, length1, length2, feedback, feedforward):
    beg: cython.long = clm.seconds2samples(time)
    end: cython.long = clm.seconds2samples(time + dur)
    s: clm.mus_any = clm.make_pulse_train(freq, amplitude=amp)
    d0: clm.mus_any = clm.make_all_pass(feedback, feedforward, size=length1, max_size=(1 + max(length1, length2)))
    zenv: clm.mus_any = clm.make_env([0,0,1,1], scaler=(length2-length1), duration=dur)
    i: cython.long = 0
    for i in range(beg, end):
        clm.outa(i, clm.all_pass(d0, clm.pulse_train(s), clm.env(zenv)))

if __name__ == '__main__': 
    with clm.Sound(play=True,statistics=True):
        za(0,1,100,.5, 20, 100, .95, .95)
        za(1.5, 1, 100, .5, 100, 20, .95, .95)



#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
if cython.compiled:
    from cython.cimports.pysndlib import clm


# --------------- touch_tone ---------------- #
# ;;; I think the dial tone is 350 + 440
# ;;; http://www.hackfaq.org/telephony/telephone-tone-frequencies.shtml

@cython.ccall
def touch_tone(start, telephone_number):
    touch_tab_1 = [0,697,697,697,770,770,770,852,852,852,941,941,941]
    touch_tab_2 = [0,1209,1336,1477,1209,1336,1477,1209,1336,1477,1209,1336,1477]
    
    
    i: cython.long = 0
    j: cython.long = 0
    
    for i in range(len(telephone_number)):
        k = telephone_number[i]
        beg: cython.long = clm.seconds2samples(start + (i * .4))
        end:  cython.long = beg + clm.seconds2samples(.3)
        if clm.is_number(k):
            i = k if not (0 == k) else 11
        else:
            i = 10 if k == '*' else 12
        
        frq1: clm.mus_any  = clm.make_oscil(touch_tab_1[i])
        frq2: clm.mus_any  = clm.make_oscil(touch_tab_2[i])
        
        for j in range(beg, end):
            clm.outa(j, (.1 * (clm.oscil(frq1) + clm.oscil(frq2))))

if __name__ == '__main__':             
    with clm.Sound( play = True, statistics=True):
        touch_tone(0, [8,6,7,5,3,0,9])
        



# cython: c_string_type=unicode
# cython: c_string_encoding=utf8

cimport cython
import numpy as np
cimport numpy as np
cimport pysndlib.cclm as cclm
cimport pysndlib.csndlib as csndlib
from pysndlib.sndlib cimport Sample, Header

np.import_array()

# --------------- clm enums ---------------- #
cpdef enum Interp:
    """
    various interpolation types
    """
    NONE, LINEAR,SINUSOIDAL, ALL_PASS, LAGRANGE, BEZIER, HERMITE

cpdef enum Window:
    """
    many useful windows
    """
    RECTANGULAR, HANN, WELCH, PARZEN, BARTLETT, HAMMING, BLACKMAN2, BLACKMAN3, BLACKMAN4, EXPONENTIAL, RIEMANN, KAISER, CAUCHY, POISSON, GAUSSIAN, TUKEY, DOLPH_CHEBYSHEV, HANN_POISSON, CONNES, SAMARAKI, ULTRASPHERICAL, BARTLETT_HANN, BOHMAN, FLAT_TOP, BLACKMAN5, BLACKMAN6, BLACKMAN7, BLACKMAN8, BLACKMAN9, BLACKMAN10, RV2, RV3, RV4, MLT_SINE, PAPOULIS, DPSS, SINC,

cpdef enum Spectrum:
    """
    types of normalizations when using the spectrum function. the results are in db if in_db, or linear and normalized to 1.0 normalized, or linear unnormalized raw
    """
    IN_DB, NORMALIZED, RAW
    
cpdef enum Polynomial:
    """
    used for polynomial based gens 
    """
    EITHER_KIND, FIRST_KIND, SECOND_KIND, BOTH_KINDS



cdef class mus_any:
    """
    a wrapper class for mus_any pointers in c
    
    instances of this class have the following properties, which may be 
    used if supported by the type of generator used to make the instance:
    
    
    """
    
    cdef cclm.mus_any *_ptr
    cdef bint ptr_owner
    cdef cclm.input_cb _inputcallback
    cdef cclm.edit_cb _editcallback
    cdef cclm.analyze_cb _analyzecallback
    cdef cclm.synthesize_cb _synthesizecallback
    cdef list _cache
    cdef cclm.mus_float_t* _data_ptr
    cdef np.ndarray _data
    cdef cclm.mus_float_t* _xcoeffs_ptr
    cdef np.ndarray _xcoeffs
    cdef cclm.mus_float_t* _ycoeffs_ptr
    cdef np.ndarray _ycoeffs
    cdef cclm.mus_float_t* _pv_amp_increments_ptr
    cdef np.ndarray _pv_amp_increments
    cdef cclm.mus_float_t* _pv_amps_ptr
    cdef np.ndarray _pv_amps
    cdef cclm.mus_float_t* _pv_freqs_ptr
    cdef np.ndarray _pv_freqs
    cdef cclm.mus_float_t* _pv_phases_ptr
    cdef np.ndarray _pv_phases
    cdef cclm.mus_float_t* _pv_phase_increments_ptr
    cdef np.ndarray _pv_phase_increments


    # these should all be considered private
    @staticmethod
    cdef mus_any from_ptr(cclm.mus_any *_ptr, bint owner=*, cython.int length=*)

    cpdef cache_append(self, obj)
    cpdef cache_extend(self, obj)
    cpdef set_up_data(self , cython.int length=*)
    cpdef set_up_xcoeffs(self)
    cpdef set_up_ycoeffs(self)
    cpdef set_up_pv_data(self)
    
    # these should all be considered public
    cpdef mus_reset(self)
    cpdef next(self, cython.double arg1=*, cython.double arg2=*)



cpdef int mus_close(mus_any obj)
cpdef bint mus_is_output(mus_any obj)
cpdef bint mus_is_input(mus_any obj)
cpdef mus_reset(mus_any obj)
cpdef cython.long get_length(obj)
cpdef cython.double get_srate(obj=*)
cpdef cython.long get_framples(obj)
cpdef cython.long get_channels(obj)
cpdef cython.double random(cython.double x=*)
cpdef bint is_zero( cython.double x)
cpdef bint is_even(n)
cpdef bint is_odd(n)
cpdef cython.double radians2hz(cython.double radians )
cpdef cython.double hz2radians(cython.double hz)
cpdef cython.double degrees2radians(cython.double degrees)
cpdef cython.double radians2degrees(cython.double radians)
cpdef cython.double db2linear(cython.double x)
cpdef cython.double linear2db(cython.double x)
cpdef cython.double odd_multiple(cython.double x, cython.double y)
cpdef cython.double even_multiple(cython.double x, cython.double y)
cpdef cython.double odd_weight(cython.double x)
cpdef cython.double even_weight(cython.double x)
cpdef cython.double set_srate(cython.double r)
cpdef cython.long seconds2samples(cython.double secs)
cpdef cython.double samples2seconds(cython.long samples)
cpdef cython.double ring_modulate(cython.double s1, cython.double s2)
cpdef cython.double amplitude_modulate(cython.double s1, cython.double s2, cython.double s3)
cpdef cython.double contrast_enhancement(cython.double insig, cython.double fm_index=* )
cpdef cython.double dot_product(np.ndarray data1, np.ndarray  data2)
cpdef cython.double polynomial(np.ndarray coeffs, cython.double x )
cpdef cython.double array_interp(np.ndarray  fn, cython.double  x)
cpdef cython.double bessi0(cython.double x)
cpdef cython.double mus_interpolate(Interp interp_type, cython.double x, np.ndarray table, cython.double y1 =*)
cpdef np.ndarray make_fft_window(Window window_type, cython.int size, cython.double beta=* , cython.double alpha=*)

cpdef file2ndarray(str filename, channel=*, beg=*, size=*)
cpdef ndarray2file(str filename, np.ndarray arr, size=*, sr=*, sample_type=*, header_type=*, comment=* )


cpdef np.ndarray mus_fft(np.ndarray rdat, np.ndarray  idat, cython.int fft_size , cython.int sign)
cpdef np.ndarray fft(np.ndarray rdat, np.ndarray idat,  cython.int fft_size, cython.int sign )
cpdef np.ndarray mus_rectangular2polar(np.ndarray  rdat, np.ndarray  idat)
cpdef tuple rectangular2polar(np.ndarray rdat, np.ndarray  idat)
cpdef np.ndarray mus_rectangular2magnitudes(np.ndarray rdat, np.ndarray idat)
cpdef tuple rectangular2magnitudes(np.ndarray rdat, np.ndarray idat)
cpdef np.ndarray mus_polar2rectangular(np.ndarray rdat,  np.ndarray idat)
cpdef tuple polar2rectangular(np.ndarray rdat, np.ndarray idat)
cpdef np.ndarray mus_spectrum(np.ndarray rdat, np.ndarray  idat, np.ndarray  window, Spectrum norm_type=*)
cpdef tuple spectrum(np.ndarray rdat,np.ndarray idat,np.ndarray window, Spectrum norm_type=*)
cpdef np.ndarray mus_spectrum(np.ndarray rdat, np.ndarray  idat, np.ndarray  window, Spectrum norm_type=*)
cpdef np.ndarray convolution(np.ndarray rl1, np.ndarray  rl2, fft_size = *)
cpdef np.ndarray mus_convolution(np.ndarray  rl1, np.ndarray  rl2, fft_size=*)
cpdef np.ndarray autocorrelate(np.ndarray data)
cpdef np.ndarray mus_autocorrelate(np.ndarray data)
cpdef np.ndarray mus_correlate( np.ndarray data1, np.ndarray  data2)
cpdef np.ndarray correlate(np.ndarray data1, np.ndarray  data2)
cpdef np.ndarray mus_cepstrum( np.ndarray data)
cpdef np.ndarray cepstrum(np.ndarray data)
cpdef np.ndarray partials2wave(partials, np.ndarray wave=*, table_size=*, bint norm=* )
cpdef np.ndarray phase_partials2wave(partials, np.ndarray wave=*, table_size=*, bint norm=* )
cpdef np.ndarray partials2polynomial(partials, Polynomial kind=*)
cpdef np.ndarray normalize_partials(partials)
cpdef cython.double chebyshev_tu_sum(cython.double x, tcoeffs, ucoeffs)
cpdef cython.double chebyshev_t_sum(cython.double x, tcoeffs)
cpdef cython.double chebyshev_u_sum(cython.double x, ucoeffs)

# ---------------- clm gen ---------------- #

cpdef cython.double oscil(mus_any gen, cython.double fm=*, cython.double pm=*)
cpdef cython.double oscil_unmodulated(mus_any gen)
cpdef bint is_oscil(mus_any gen)


cpdef cython.double oscil_bank(mus_any gen)
cpdef bint is_oscil_bank(mus_any gen)

#
cpdef cython.double env(mus_any gen )
cpdef bint is_env(mus_any gen)
cpdef cython.double env_interp(cython.double x , mus_any env)
cpdef cython.double envelope_interp(cython.double x, mus_any env)
cpdef np.ndarray env_rates( mus_any gen)
cpdef cython.double env_any(mus_any gen , connection_function)

cpdef cython.double pulsed_env(mus_any gen, cython.double fm=*)
cpdef cython.double pulsed_env_unmodulated(mus_any gen)
cpdef bint is_pulsed_env(mus_any gen)

cpdef cython.double table_lookup(mus_any gen, cython.double fm=*)
cpdef cython.double table_lookup_unmodulated(mus_any gen)
cpdef bint is_table_lookup(mus_any gen)
cpdef mus_any make_table_lookup_with_env(cython.double frequency, envelope, size= *)

cpdef cython.double polywave(mus_any gen, cython.double fm=*)
cpdef cython.double polywave_unmodulated(mus_any gen)        
cpdef bint is_polywave(mus_any gen)

cpdef cython.double polyshape(mus_any gen, cython.double index=*, cython.double fm=*)
cpdef cython.double polyshape_unmodulated(mus_any gen, cython.double index=*)
cpdef bint is_polyshape(mus_any gen)

cpdef cython.double triangle_wave(mus_any gen, cython.double fm=*)
cpdef cython.double triangle_wave_unmodulated(mus_any gen)
cpdef bint is_triangle_wave(mus_any gen)

cpdef cython.double square_wave(mus_any gen, cython.double fm=*)
cpdef bint is_square_wave(mus_any gen)

cpdef mus_any make_sawtooth_wave(cython.double frequency, cython.double amplitude=*, cython.double phase=*)
cpdef cython.double sawtooth_wave(mus_any gen, cython.double fm =*)
cpdef bint is_sawtooth_wave(mus_any gen)

cpdef cython.double pulse_train(mus_any gen, cython.double fm=*)
cpdef cython.double pulse_train_unmodulated(mus_any gen)
cpdef bint is_pulse_train(mus_any gen)

cpdef cython.double ncos(mus_any gen, cython.double fm=*)
cpdef bint is_ncos(mus_any gen)

cpdef cython.double nsin(mus_any gen, cython.double fm=*)
cpdef bint is_nsin(mus_any gen )

cpdef cython.double nrxysin(mus_any gen, cython.double fm=*)
cpdef bint is_nrxysin(mus_any gen)

cpdef cython.double nrxycos(mus_any gen, cython.double fm=*)
cpdef bint is_nrxycos(mus_any gen)

cpdef cython.double rxykcos(mus_any gen, cython.double fm=*)
cpdef bint is_rxykcos(mus_any gen)

cpdef cython.double rxyksin(mus_any gen, cython.double fm=*)
cpdef bint is_rxyksin(mus_any gen)

cpdef cython.double ssb_am(mus_any gen, cython.double insig, cython.double fm=*)
cpdef cython.double ssb_am_unmodulated(mus_any gen, cython.double insig)
cpdef bint is_ssb_am(mus_any gen)

cpdef cython.double wave_train(mus_any gen, cython.double fm=*)
cpdef cython.double wave_train_unmodulated(mus_any gen)
cpdef bint is_wave_train(mus_any gen)

cpdef cython.double rand(mus_any gen, cython.double sweep=*)
cpdef cython.double rand_unmodulated(mus_any gen)
cpdef bint is_rand(mus_any gen)

cpdef rand_interp(mus_any gen, cython.double sweep=*)
cpdef rand_interp_unmodulated(mus_any gen)
cpdef bint is_rand_interp(mus_any gen)

cpdef cython.double one_pole(mus_any gen, cython.double insig)
cpdef bint is_one_pole(mus_any gen)

cpdef cython.double one_zero(mus_any gen, cython.double insig)
cpdef bint is_one_zero(mus_any gen)

cpdef cython.double two_pole(mus_any gen, cython.double insig)
cpdef bint is_two_pole(mus_any gen)

cpdef cython.double two_zero(mus_any gen, cython.double insig)
cpdef bint is_two_zero(mus_any gen)

cpdef cython.double formant(mus_any gen, cython.double insig, cython.double radians=*)
cpdef bint is_formant(mus_any gen)

cpdef cython.double formant_bank(mus_any gen, inputs)
cpdef bint is_formant_bank(mus_any gen)

cpdef firmant(mus_any gen, cython.double insig, cython.double radians=*)
cpdef is_firmant(mus_any gen)

cpdef cython.double filter(mus_any gen, cython.double insig)
cpdef bint is_filter(mus_any gen)

cpdef cython.double fir_filter(mus_any gen, cython.double insig)
cpdef bint is_fir_filter(mus_any gen)

cpdef cython.double iir_filter(mus_any gen, cython.double insig)
cpdef bint is_iir_filter(mus_any gen)

cpdef cython.double delay(mus_any gen, cython.double insig, cython.double pm=*)
cpdef cython.double delay_unmodulated(mus_any gen, cython.double insig)
cpdef bint is_delay(mus_any gen)
cpdef cython.double tap(mus_any gen, offset=*)
cpdef cython.double tap_unmodulated(mus_any gen)
cpdef bint is_tap(mus_any gen)
cpdef cython.double delay_tick(mus_any gen, cython.double insig )

cpdef cython.double comb(mus_any gen, cython.double insig, pm=*)
cpdef bint is_comb(mus_any gen)

cpdef cython.double comb_bank(mus_any gen, cython.double insig)
cpdef bint is_comb_bank(mus_any gen)

cpdef cython.double filtered_comb(mus_any gen, cython.double insig, cython.double pm=*)
cpdef cython.double filtered_comb_unmodulated(mus_any gen, cython.double insig)
cpdef bint is_filtered_comb(mus_any gen)

cpdef cython.double filtered_comb_bank(mus_any gen)
cpdef bint is_filtered_comb_bank(mus_any gen)

cpdef cython.double notch(mus_any gen, cython.double insig, cython.double pm=*)
cpdef cython.double notch_unmodulated(mus_any gen, cython.double insig)
cpdef bint is_notch(mus_any gen)

cpdef cython.double all_pass(mus_any gen, cython.double insig, pm=*)
cpdef is_all_pass(mus_any gen)

cpdef cython.double all_pass_bank(mus_any gen, cython.double insig)
cpdef bint is_all_pass_bank(mus_any gen)

cpdef cython.double one_pole_all_pass(mus_any gen, cython.double insig)
cpdef bint is_one_pole_all_pass(mus_any gen)

cpdef cython.double moving_average(mus_any gen, cython.double  insig)
cpdef bint is_moving_average(mus_any gen)

cpdef cython.double moving_max(mus_any gen, cython.double insig)
cpdef bint is_moving_max(mus_any gen)

cpdef cython.double moving_norm(mus_any gen, cython.double insig)
cpdef is_moving_norm(mus_any gen)

cpdef cython.double asymmetric_fm(mus_any gen, cython.double index, cython.double fm=*)
cpdef cython.double asymmetric_fm_unmodulated(mus_any gen, cython.double index)
cpdef bint is_asymmetric_fm(mus_any gen)

cpdef cython.double file2sample(mus_any gen, cython.long loc, cython.int chan=*)
cpdef bint is_file2sample(mus_any gen)

cpdef cython.double sample2file(mus_any gen, cython.long samp, cython.int chan, cython.double val)
cpdef bint is_sample2file(mus_any gen)

cpdef file2frample(mus_any gen, cython.long loc)
cpdef is_file2frample(mus_any gen)

cpdef cython.double frample2file(mus_any gen, cython.long samp, vals)
cpdef bint is_frample2file(mus_any gen)
cpdef mus_any continue_frample2file(str name)

cpdef cython.double readin(mus_any gen)
cpdef is_readin(mus_any gen)

cpdef cython.double src(mus_any gen, cython.double sr_change=*)
cpdef bint is_src(mus_any gen)

cpdef cython.double convolve(mus_any gen)
cpdef bint is_convolve(mus_any gen)

cpdef cython.double granulate(mus_any gen)
cpdef bint is_granulate(mus_any e)

cpdef cython.double phase_vocoder(mus_any gen)
cpdef bint is_phase_vocoder(mus_any gen)
cpdef np.ndarray phase_vocoder_amp_increments(mus_any gen)
cpdef np.ndarray phase_vocoder_amps(mus_any gen)
cpdef np.ndarray phase_vocoder_freqs(mus_any gen)
cpdef np.ndarray phase_vocoder_phases(mus_any gen)
cpdef np.ndarray phase_vocoder_phase_increments(mus_any gen)

cpdef out_any(cython.long loc, cython.double data, cython.int channel, output=*)
cpdef outa(cython.long loc, cython.double data, output=*)
cpdef outb(cython.long loc, cython.double data, output=*)
cpdef outc(cython.long loc, cython.double data, output=*)
cpdef outd(cython.long loc, cython.double data, output=*)
cpdef in_any(cython.long loc, cython.int channel, inp)

cpdef ina(cython.long loc, inp)
cpdef inb(cython.long loc, inp)

cpdef out_bank(gens, cython.long loc, cython.double val)

cpdef cython.double locsig(mus_any gen, cython.long loc, cython.double val)
cpdef bint is_locsig(mus_any gen)
cpdef cython.double locsig_ref(mus_any gen, cython.int chan)
cpdef cython.double locsig_set(mus_any gen, cython.int chan, cython.double val)
cpdef cython.double locsig_reverb_ref(mus_any gen, cython.int chan)
cpdef cython.double locsig_reverb_set(mus_any gen, cython.int chan, cython.double val)
cpdef move_locsig(mus_any gen,cython.double degree, cython.double distance)

cpdef convolve_files(str file1, str file2, cython.double maxamp=*, str outputfile=*, sample_type=*, header_type=*)


#cpdef mus_any make_dcblock()
cpdef cython.double dcblock(mus_any gen, cython.double insig)

cpdef cython.double biquad(mus_any gen, cython.double insig)
cpdef bint is_biquad(mus_any gen)
cpdef biquad_set_resonance(mus_any gen, cython.double fc, cython.double radius, bint normalize=*)
cpdef biquad_set_equal_gain_zeroes(mus_any gen)
cpdef cython.double bes_j0(cython.double x)
cpdef cython.double bes_j1(cython.double x)
cpdef cython.double bes_jn(cython.int n, cython.double x)
cpdef cython.double bes_y0(cython.double x)  
cpdef cython.double bes_y1(cython.double x)
cpdef cython.double bes_yn(cython.int n, cython.double x)


cpdef cython.doublecomplex edot_product(cython.doublecomplex freq, np.ndarray data, size=*)

#cpdef mus_any make_oscil(cython.double frequency=*,cython.double  initial_phase = *)
#cpdef mus_any make_oscil_bank(freqs, phases, amps=*, bint stable=*)
#cpdef mus_any make_env(envelope, cython.double duration=*, cython.double scaler=*, cython.double offset=*, cython.double base=*, cython.int length=*)
#cpdef mus_any make_pulsed_env(envelope, cython.double duration , cython.double frequency )
#cpdef mus_any make_table_lookup(cython.double frequency=*, cython.double initial_phase=*, wave=*, cython.int size=*, interp interp_type=*)
#cpdef mus_any make_polywave(cython.double frequency, partials = *, polynomial kind=*, xcoeffs=*, ycoeffs=*)
#cpdef mus_any make_polyshape(cython.double frequency, cython.double initial_phase=*, coeffs = *, partials = *, polynomial kind=*)
#cpdef mus_any make_triangle_wave(cython.double frequency, cython.double amplitude=*, cython.double phase=*)
#cpdef mus_any make_square_wave(cython.double frequency , cython.double amplitude=*, cython.double phase=*)
#cpdef mus_any make_pulse_train(cython.double frequency=*, cython.double amplitude=*, cython.double phase=*)
#cpdef mus_any make_ncos(cython.double frequency, cython.int n=*)
#cpdef mus_any make_nsin(cython.double frequency, cython.int n =*)
#cpdef mus_any make_nrxysin( cython.double frequency,  cython.double ratio=*, cython.int n=*, cython.double r=*)
#cpdef mus_any make_nrxycos(cython.double frequency , cython.double ratio=*, cython.int n=*, cython.double r=*)
#cpdef mus_any make_rxykcos(cython.double frequency, cython.double phase=*, cython.double r=*, cython.double ratio=*)
#cpdef mus_any make_rxyksin(cython.double frequency, cython.double phase, cython.double r=*, cython.double ratio=*)
#cpdef mus_any make_ssb_am(cython.double frequency, cython.int order=*)
#cpdef mus_any make_wave_train(cython.double frequency, wave=*, cython.double initial_phase=*, interp interp_type=*)
#cpdef mus_any make_wave_train_with_env(cython.double frequency, envelope, size=*)
#cpdef mus_any make_rand(cython.double frequency, cython.double amplitude=*, distribution=*)
#cpdef mus_any make_rand_interp(cython.double frequency, cython.double amplitude=*, distribution=*)
#cpdef mus_any make_one_pole(cython.double a0, cython.double b1)
#cpdef mus_any make_one_zero(cython.double a0 , cython.double a1)
#cpdef make_two_pole(frequency=*, radius=*, a0=*,b1=*,b2=*)
#cpdef make_two_zero(frequency=*, radius=*, a0=*,a1=*,a2=*)
#cpdef mus_any make_formant(cython.double frequency, cython.double radius)
#cpdef mus_any make_formant_bank(list filters, amps=*)
#cpdef mus_any make_firmant(cython.double frequency, cython.double radius)
#cpdef mus_any make_filter(cython.int order, xcoeffs, ycoeffs)
#cpdef mus_any make_fir_filter(cython.int order, xcoeffs)
#cpdef mus_any make_iir_filter(cython.int order, ycoeffs)
#cpdef np.ndarray make_fir_coeffs(cython.int order, envelope)
#cpdef mus_any make_delay(cython.long size, initial_contents=*, initial_element=*, max_size=*, interp interp_type=*)
#cpdef mus_any make_comb(cython.double feedback=*,size=*, initial_contents=*, initial_element=*, max_size=*, interp interp_type=*)
#cpdef mus_any make_comb_bank(list combs)
#cpdef mus_any make_filtered_comb(cython.double feedback,cython.long size, mus_any filter, initial_contents=*,  cython.double initial_element=*, max_size=*, interp interp_type=*)
#cpdef mus_any make_filtered_comb_bank(list fcombs)
#cpdef mus_any make_notch(cython.double feedforward=*,size=*, initial_contents=*, cython.double initial_element=*, max_size=*,interp interp_type=*)
#cpdef mus_any make_all_pass(cython.double feedback, cython.double  feedforward, cython.long  size, initial_contents= *, cython.double  initial_element = *,  max_size= *, interp interp_type = *)
#cpdef mus_any make_all_pass_bank(list all_passes)
#cpdef mus_any make_one_pole_all_pass(cython.long size, cython.double coeff)
#cpdef mus_any make_moving_average(cython.long size, initial_contents=*, cython.double initial_element=*)
#cpdef mus_any make_moving_max(cython.long size, initial_contents=*, cython.double initial_element=*)
#cpdef mus_any make_moving_norm(cython.long size, initial_contents=*, cython.double scaler=*)
#cpdef mus_any make_asymmetric_fm(cython.double frequency, cython.double initial_phase=*, cython.double r=*, cython.double ratio=*)
#cpdef mus_any make_file2sample(str filename, buffer_size=*)
#cpdef mus_any make_sample2file(str filename, cython.int chans=*, sample sample_type=*, header header_type=*, comment=*)
#cpdef mus_any make_file2frample(str filename, buffer_size=*)
#cpdef mus_any make_frample2file(str filename, cython.int chans=*, sample sample_type=*, header header_type=*, comment=*)
#cpdef mus_any make_readin(str filename, cython.int chan=*, cython.long start=*, cython.int direction=*, buffer_size=*)
#cpdef mus_any make_src(inp , cython.double srate=*, cython.int width=*)
#cpdef mus_any make_convolve(inp, filt, cython.long fft_size=*, filter_size=*)
#cpdef mus_any make_granulate(inp, cython.double expansion=*, cython.double length=*, cython.double scaler=*, cython.double hop=*, cython.double ramp=*, cython.double jitter=*, cython.int max_size=*, edit=*)
#cpdef make_phase_vocoder(inp, cython.int fft_size=*, cython.int overlap=*, cython.int  interp=*, cython.double pitch=*, analyze=*, edit=*, synthesize=*)
#cpdef mus_any make_locsig(cython.double degree=*, cython.double distance=*,  cython.double reverb=*, output=*, revout=*, channels=*, reverb_channels=*,interp interp_type=*)






#==================================================================================
# The code is ported from Bill Schottstedaet's 'generators.scm' 
# file available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
import pysndlib.generators as gens
if cython.compiled:
    from cython.cimports.pysndlib import clm
    from cython.cimports.pysndlib import generators as gens

# from generators.scm 
# uses fmssb generator

# --------------- machine1 ---------------- #
@cython.ccall
def machine1(start, dur, cfreq, mfreq, amp, index, gliss):
    gen = gens.make_fmssb(cfreq, mfreq/cfreq, index=1.)
    beg: cython.long = clm.seconds2samples(start)
    end: cython.long = clm.seconds2samples(start+dur)
    ampf = clm.make_env([0,0,1,.75,2,1,3,.1,4,.7,5,1,6,.8,100,0], base=32, scaler=amp, duration=dur)
    indf = clm.make_env([0,0,1,1,3,0], duration=dur, base=32, scaler=index)
    frqf = clm.make_env([0,0,1,1] if gliss > 0.0 else [0,1,1,0], duration=dur, scaler=clm.hz2radians((cfreq/mfreq) * math.fabs(gliss)))
    i: cython.long = 0
    for i in range(beg, end):
        gen.index = clm.env(indf)
        clm.outa(i, clm.env(ampf) * gens.fmssb(gen, clm.env(frqf)))


if __name__ == '__main__': 
    with clm.Sound(play=True,statistics=True):
        for i in np.arange(0.0, 2.0, .5):        
            machine1(i, .3, 100, 540, 0.5, 3.0, 0.0)
            machine1(i,.1,100,1200,.5,10.0,200.0)
            machine1(i,.3,100,50,.75,10.0,0.0)
            machine1(i + .1, .1,100,1200,.5,20.0,1200.0)
            machine1(i + .3, .1,100,1200,.5,20.0,1200.0)
            machine1(i + .3, .1,100,200,.5,10.0,200.0)
            machine1(i + .36, .1,100,200,.5,10.0,200.0)
            machine1(i + .4, .1,400,300,.5,10.0,-900.0)
            machine1(i + .4, .21,100,50,.75,10.0,1000.0)

        for i in np.arange(3, 5, .2):
            machine1(i, .3, 100, 540, .5, 4.0, 0.)
            machine1(i+.1, .3, 200, 540, .5, 3.0, 0.0)

        for i in np.arange(5., 7., .6):
            machine1(i, .3, 1000, 540, .5, 6., 0.)
            machine1(i+.1, .1, 2000, 540, .5, 1.0, 0)
